from __future__ import annotations

import copy
import io
import logging
import os
from pathlib import Path
from tkinter import E
from typing import Optional, Union

from ruamel.yaml import YAML

from hpcflow.sdk.core.validation import get_schema

from .errors import (
    ConfigChangeFileUpdateError,
    ConfigDefaultValidationError,
    ConfigFileInvocationIncompatibleError,
    ConfigFileValidationError,
    ConfigInvocationKeyNotFoundError,
    ConfigValidationError,
)


class ConfigFile:
    """Configuration file."""

    def __init__(self, config, directory, invoc_key=None):

        self.config = config
        self.logger = self.config._logger
        self.directory = self._resolve_config_dir(directory)

        # set by _load_file_data:
        self.path = None
        self.contents = None
        self.data = None
        self.data_rt = None

        self._load_file_data()
        self._validate(self.data)

        # select appropriate config data for this invocation using run-time info:
        if not invoc_key:
            for c_name_i, c_dat_i in self.data["configs"].items():
                is_match = True
                for match_k, match_v in c_dat_i["invocation"]["match"].items():
                    if getattr(self.config._app.run_time_info, match_k) != match_v:
                        is_match = False
                        break
                if is_match:
                    invoc_key = c_name_i
                    self.logger.debug(
                        f"Found matching config ({invoc_key!r}) for this invocation."
                    )
                    break

            if not is_match:
                raise ConfigFileInvocationIncompatibleError(invoc_key)

        elif invoc_key not in self.data["configs"]:
            raise ConfigInvocationKeyNotFoundError(
                invoc_key,
                self.path,
                list(self.data["configs"].keys()),
            )

        self.invoc_key = invoc_key

    def _validate(self, data):
        file_schema = get_schema("config_file_schema.yaml")
        file_validated = file_schema.validate(data)
        if not file_validated.is_valid:
            raise ConfigFileValidationError(file_validated.get_failures_string())

    @property
    def invoc_data(self):
        return self.data["configs"][self.invoc_key]

    def save(self):

        new_data = copy.deepcopy(self.data)
        new_data_rt = copy.deepcopy(self.data_rt)
        new_contents = ""

        modified_names = list(self.config._modified_keys.keys()) + self.config._unset_keys
        for k, v in self.config._modified_keys.items():
            new_data["configs"][self.invoc_key]["config"][k] = v
            new_data_rt["configs"][self.invoc_key]["config"][k] = v

        for k in self.config._unset_keys:
            del new_data["configs"][self.invoc_key]["config"][k]
            del new_data_rt["configs"][self.invoc_key]["config"][k]

        try:
            # write a new temporary config file
            cfg_file_path = self.config.get("config_file_path")
            cfg_tmp_file = cfg_file_path.with_suffix(cfg_file_path.suffix + ".tmp")
            self.logger.debug(f"Creating temporary config file: {cfg_tmp_file!r}.")
            yaml = YAML(typ="rt")
            with cfg_tmp_file.open("wt", newline="\n") as fh:
                yaml.dump(new_data_rt, fh)

            # atomic rename, overwriting original:
            self.logger.debug("Replacing original config file with temporary file.")
            os.replace(src=cfg_tmp_file, dst=cfg_file_path)

            buff = io.BytesIO()
            yaml.dump(new_data_rt, buff)
            new_contents = buff.getvalue()

        except Exception as err:
            raise ConfigChangeFileUpdateError(names=modified_names, err=err) from None

        self.data = new_data
        self.data_rt = new_data_rt
        self.contents = new_contents

        self.config._unset_keys = []
        self.config._modified_keys = {}

    def _resolve_config_dir(self, directory: Optional[Union[str, Path]] = None) -> Path:
        """Find the directory in which to locate the configuration file.

        If no configuration directory is specified, look first for an environment variable
        (given by config option `directory_env_var`), and then in the default
        configuration directory (given by config option `default_directory`).

        The configuration directory will be created if it does not exist.

        Parameters
        ----------
        directory
            Directory in which to find the configuration file. Optional.

        Returns
        -------
        directory : Path
            Absolute path to the configuration directory.

        """

        if not directory:
            directory = Path(
                os.getenv(
                    self.config._options.directory_env_var,
                    self.config._options.default_directory,
                )
            ).expanduser()
        else:
            directory = Path(directory)

        if not directory.is_dir():
            self.logger.debug(
                f"Configuration directory does not exist. Generating here: {str(directory)!r}."
            )
            directory.mkdir()
        else:
            self.logger.debug(f"Using configuration directory: {str(directory)!r}.")

        return directory.resolve()

    def _setup_default_config(self, config_path):
        self.config._logger.info(
            "No config.yaml found in the configuration directory. Generating "
            "a config.yaml file."
        )

        # validate the default:
        default_config = self.config._options.default_config
        try:
            # validate default config "file" structure:
            self._validate(data=default_config)

            # validate default config items for each defined invocation:
            for val in default_config["configs"].values():
                self.config._validate(data=val["config"], raise_with_metadata=False)

        except (ConfigFileValidationError, ConfigValidationError) as err:
            raise ConfigDefaultValidationError(err) from None

        yaml = YAML(typ="rt")
        with config_path.open("w") as handle:
            yaml.dump(default_config, handle)

    def _load_file_data(self):
        """Load data from the configuration file (config.yaml or config.yml)."""

        # Try both ".yml" and ".yaml" extensions:
        path = self.directory.joinpath("config.yaml")
        if not path.is_file():
            path = self.directory.joinpath("config.yml")

        if not path.is_file():
            self._setup_default_config(config_path=path)

        yaml = YAML(typ="safe")
        yaml_rt = YAML(typ="rt")
        with path.open() as handle:
            contents = handle.read()
            handle.seek(0)
            data = yaml.load(handle)
            handle.seek(0)
            data_rt = yaml_rt.load(handle)

        self.path = path
        self.contents = contents
        self.data = data
        self.data_rt = data_rt

    def get_config_item(self, name, raise_on_missing=False, default_value=None):
        if raise_on_missing and name not in self.invoc_data["config"]:
            raise ValueError(f"missing from file: {name!r}")
        return self.invoc_data["config"].get(name, default_value)

    def is_item_set(self, name):
        try:
            self.get_config_item(name, raise_on_missing=True)
        except ValueError:
            return False
        return True



from __future__ import annotations
import copy
from dataclasses import dataclass
import enum
import re
import subprocess
from typing import Any, Dict, List, Optional, Tuple, Union

from valida.rules import Rule

from hpcflow.sdk.core.command_files import FileSpec, InputFileGenerator, OutputFileParser
from hpcflow.sdk.core.commands import Command
from hpcflow.sdk.core.environment import Environment
from hpcflow.sdk.core.errors import MissingCompatibleActionEnvironment
from hpcflow.sdk.core.json_like import ChildObjectSpec, JSONLike


ACTION_SCOPE_REGEX = r"(\w*)(?:\[(.*)\])?"


@dataclass(eq=True, frozen=True)
class ElementID:
    task_insert_ID: int
    element_idx: int

    def __lt__(self, other):
        return tuple(self.__dict__.values()) < tuple(other.__dict__.values())


@dataclass(eq=True, frozen=True)
class IterationID(ElementID):
    """
    Attributes
    ----------
    iteration_idx :
        Index into the `element_iterations` list/array of the task. Note this is NOT the
        index into local list of ElementIterations belong to an Element.
    """

    iteration_idx: int

    def get_element_ID(self):
        return ElementID(
            task_insert_ID=self.task_insert_ID,
            element_idx=self.element_idx,
        )


@dataclass(eq=True, frozen=True)
class EAR_ID(IterationID):
    action_idx: int
    run_idx: int
    EAR_idx: int

    def get_iteration_ID(self):
        return IterationID(
            task_insert_ID=self.task_insert_ID,
            element_idx=self.element_idx,
            iteration_idx=self.iteration_idx,
        )


class ActionScopeType(enum.Enum):

    ANY = 0
    MAIN = 1
    PROCESSING = 2
    INPUT_FILE_GENERATOR = 3
    OUTPUT_FILE_PARSER = 4


ACTION_SCOPE_ALLOWED_KWARGS = {
    ActionScopeType.ANY.name: set(),
    ActionScopeType.MAIN.name: set(),
    ActionScopeType.PROCESSING.name: set(),
    ActionScopeType.INPUT_FILE_GENERATOR.name: {"file"},
    ActionScopeType.OUTPUT_FILE_PARSER.name: {"output"},
}


class EARSubmissionStatus(enum.Enum):

    PENDING = 0  # Not yet associated with a submission
    PREPARED = 1  # Associated with a submission that is not yet submitted
    SUBMITTED = 2  # Submitted for execution
    RUNNING = 3  # Executing
    COMPLETE = 4  # Finished executing


class ElementActionRun:
    _app_attr = "app"

    def __init__(
        self, element_action, run_idx: int, index: int, data_idx: Dict, metadata: Dict
    ) -> None:
        self._element_action = element_action
        self._run_idx = run_idx  # local index of this run with the action
        self._index = index  # task-wide EAR index
        self._data_idx = data_idx
        self._metadata = metadata

        # assigned on first access of corresponding properties:
        self._inputs = None
        self._outputs = None
        self._resources = None
        self._input_files = None
        self._output_files = None

    @property
    def element_action(self):
        return self._element_action

    @property
    def action(self):
        return self.element_action.action

    @property
    def element_iteration(self):
        return self.element_action.element_iteration

    @property
    def element(self):
        return self.element_iteration.element

    @property
    def workflow(self):
        return self.element_iteration.workflow

    @property
    def run_idx(self):
        return self._run_idx

    @property
    def index(self):
        """Task-wide EAR index."""
        return self._index

    @property
    def EAR_ID(self):
        """EAR index object."""
        return EAR_ID(
            EAR_idx=self.index,
            task_insert_ID=self.task.insert_ID,
            element_idx=self.element.index,
            iteration_idx=self.element_iteration.index,
            action_idx=self.element_action.action_idx,
            run_idx=self.run_idx,
        )

    @property
    def data_idx(self):
        return self._data_idx

    @property
    def metadata(self):
        return self._metadata

    @property
    def submission_idx(self):
        return self.metadata["submission_idx"]

    @property
    def start_time(self):
        return self.metadata["start_time"]

    @property
    def end_time(self):
        return self.metadata["end_time"]

    @property
    def success(self):
        return self.metadata["success"]

    @property
    def task(self):
        return self.element_action.task

    @property
    def submission_status(self):

        if self.metadata["end_time"] is not None:
            return EARSubmissionStatus.COMPLETE

        elif self.metadata["start_time"] is not None:
            return EARSubmissionStatus.RUNNING

        elif self.submission_idx is not None:

            wk_sub_stat = self.workflow.submissions[self.submission_idx].status

            if wk_sub_stat.name == "PENDING":
                return EARSubmissionStatus.PREPARED

            elif wk_sub_stat.name == "SUBMITTED":
                return EARSubmissionStatus.SUBMITTED

            else:
                RuntimeError(f"Workflow submission status not understood: {wk_sub_stat}.")

        return EARSubmissionStatus.PENDING

    def get_parameter_names(self, prefix):
        return self.element_action.get_parameter_names(prefix)

    def get_data_idx(self, path: str = None):
        return self.element_iteration.get_data_idx(
            path,
            action_idx=self.element_action.action_idx,
            run_idx=self.run_idx,
        )

    def get_parameter_sources(
        self,
        path: str = None,
        typ: str = None,
        as_strings: bool = False,
        use_task_index: bool = False,
    ):
        return self.element_iteration.get_parameter_sources(
            path,
            action_idx=self.element_action.action_idx,
            run_idx=self.run_idx,
            typ=typ,
            as_strings=as_strings,
            use_task_index=use_task_index,
        )

    def get(
        self,
        path: str = None,
        default: Any = None,
        raise_on_missing: bool = False,
    ):
        return self.element_iteration.get(
            path=path,
            action_idx=self.element_action.action_idx,
            run_idx=self.run_idx,
            default=default,
            raise_on_missing=raise_on_missing,
        )

    def get_EAR_dependencies(self, as_objects=False):
        """Get EARs that this EAR depends on."""

        out = []
        for src in self.get_parameter_sources(typ="EAR_output").values():
            src = copy.deepcopy(src)
            src.pop("type")
            _EAR_ID = EAR_ID(**src)
            if _EAR_ID != self.EAR_ID:
                # don't record a self dependency!
                out.append(_EAR_ID)

        out = sorted(out)

        if as_objects:
            out = self.workflow.get_EARs_from_indices(out)

        return out

    def get_input_dependencies(self):
        """Get information about locally defined input, sequence, and schema-default
        values that this EAR depends on. Note this does not get values from this EAR's
        task/schema, because the aim of this method is to help determine which upstream
        tasks this EAR depends on."""

        out = {}
        for k, v in self.get_parameter_sources().items():
            if (
                v["type"] in ["local_input", "default_input"]
                and v["task_insert_ID"] != self.task.insert_ID
            ):
                out[k] = v

        return out

    @property
    def inputs(self):
        if not self._inputs:
            self._inputs = self.app.ElementInputs(element_action_run=self)
        return self._inputs

    @property
    def outputs(self):
        if not self._outputs:
            self._outputs = self.app.ElementOutputs(element_action_run=self)
        return self._outputs

    @property
    def resources(self):
        if not self._resources:
            self._resources = self.app.ElementResources(**self.get_resources())
        return self._resources

    @property
    def input_files(self):
        if not self._input_files:
            self._input_files = self.app.ElementInputFiles(element_action_run=self)
        return self._input_files

    @property
    def output_files(self):
        if not self._output_files:
            self._output_files = self.app.ElementOutputFiles(element_action_run=self)
        return self._output_files

    def get_template_resources(self):
        """Get template-level resources."""
        out = {}
        for res_i in self.workflow.template.resources:
            out[res_i.scope.to_string()] = res_i._get_value()
        return out

    def get_resources(self):
        """Resolve specific resources for this EAR, considering all applicable scopes and
        template-level resources."""

        resource_specs = copy.deepcopy(self.get("resources"))
        template_resource_specs = copy.deepcopy(self.get_template_resources())
        resources = {}
        for scope in self.action.get_possible_scopes()[::-1]:
            # loop in reverse so higher-specificity scopes take precedence:
            scope_s = scope.to_string()
            if scope_s in resource_specs:
                scope_res = resource_specs[scope_s]
                if scope_s in template_resource_specs:
                    for k, v in template_resource_specs[scope_s].items():
                        if scope_res.get(k) is None and v is not None:
                            scope_res[k] = v

                resources.update({k: v for k, v in scope_res.items() if v is not None})

        return resources

    def compose_commands(self, jobscript: Jobscript) -> Tuple[str, List[str]]:
        """
        Returns
        -------
        commands
        shell_vars
            List of shell variable names that must be saved as workflow parameter data
            as strings.
        """

        # TODO: this is bash-specific

        vars_regex = r"(\<\<parameter:{}\>\>?)"
        command_lns = []
        shell_vars = []
        for command in self.action.commands:

            # substitute input parameters in command:
            cmd_str = command.command
            for cmd_inp in self.action.get_command_input_types():
                inp_val = self.get(f"inputs.{cmd_inp}")
                cmd_str = re.sub(
                    pattern=vars_regex.format(cmd_inp),
                    repl=str(inp_val),
                    string=cmd_str,
                )

            out_types = command.get_output_types()
            if out_types["stdout"]:
                # assign stdout to a shell variable if required:
                param_name = f"outputs.{out_types['stdout']}"
                shell_var_name = f"parameter_{out_types['stdout']}"
                shell_vars.append((param_name, shell_var_name))
                cmd_str = jobscript.shell.format_stream_assignment(
                    shell_var_name=shell_var_name,
                    command=cmd_str,
                )

            # TODO: also map stderr/both if possible

            command_lns.append(cmd_str)

        commands = "\n".join(command_lns) + "\n"

        return commands, shell_vars


class ElementAction:

    _app_attr = "app"

    def __init__(self, element_iteration, action_idx, runs):
        self._element_iteration = element_iteration
        self._action_idx = action_idx
        self._runs = runs

        # assigned on first access of corresponding properties:
        self._run_objs = None
        self._inputs = None
        self._outputs = None
        self._resources = None
        self._input_files = None
        self._output_files = None

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"scope={self.action.get_precise_scope().to_string()!r}, "
            f"action_idx={self.action_idx}, num_runs={self.num_runs}"
            f")"
        )

    @property
    def element_iteration(self):
        return self._element_iteration

    @property
    def element(self):
        return self.element_iteration.element

    @property
    def num_runs(self):
        return len(self._runs)

    @property
    def runs(self):
        if self._run_objs is None:
            self._run_objs = [
                self.app.ElementActionRun(self, run_idx=run_idx, **i)
                for run_idx, i in enumerate(self._runs)
            ]
        return self._run_objs

    @property
    def task(self):
        return self.element_iteration.task

    @property
    def action_idx(self):
        return self._action_idx

    @property
    def action(self):
        return self.task.template.get_schema_action(self.action_idx)

    @property
    def inputs(self):
        if not self._inputs:
            self._inputs = self.app.ElementInputs(element_action=self)
        return self._inputs

    @property
    def outputs(self):
        if not self._outputs:
            self._outputs = self.app.ElementOutputs(element_action=self)
        return self._outputs

    @property
    def input_files(self):
        if not self._input_files:
            self._input_files = self.app.ElementInputFiles(element_action=self)
        return self._input_files

    @property
    def output_files(self):
        if not self._output_files:
            self._output_files = self.app.ElementOutputFiles(element_action=self)
        return self._output_files

    def get_data_idx(self, path: str = None, run_idx: int = -1):
        return self.element_iteration.get_data_idx(
            path,
            action_idx=self.action_idx,
            run_idx=run_idx,
        )

    def get_parameter_sources(
        self,
        path: str = None,
        run_idx: int = -1,
        typ: str = None,
        as_strings: bool = False,
        use_task_index: bool = False,
    ):
        return self.element_iteration.get_parameter_sources(
            path,
            action_idx=self.action_idx,
            run_idx=run_idx,
            typ=typ,
            as_strings=as_strings,
            use_task_index=use_task_index,
        )

    def get(
        self,
        path: str = None,
        run_idx: int = -1,
        default: Any = None,
        raise_on_missing: bool = False,
    ):
        return self.element_iteration.get(
            path=path,
            action_idx=self.action_idx,
            run_idx=run_idx,
            default=default,
            raise_on_missing=raise_on_missing,
        )

    def get_parameter_names(self, prefix):
        if prefix == "inputs":
            return list(f"{i}" for i in self.action.get_input_types())
        elif prefix == "outputs":
            return list(f"{i}" for i in self.action.get_output_types())
        elif prefix == "input_files":
            return list(f"{i}" for i in self.action.get_input_file_labels())
        elif prefix == "output_files":
            return list(f"{i}" for i in self.action.get_output_file_labels())


@dataclass
class ElementActionOLD:

    _app_attr = "app"

    element: Element
    root_action: Action
    commands: List[Command]

    input_file_generator: Optional[InputFileGenerator] = None
    output_parser: Optional[OutputFileParser] = None

    def get_environment(self):
        # TODO: select correct environment according to scope:
        return self.root_action.environments[0].environment

    def execute(self):
        vars_regex = r"\<\<(executable|parameter|script|file):(.*?)\>\>"
        env = None
        resolved_commands = []
        scripts = []
        for command in self.commands:

            command_resolved = command.command
            re_groups = re.findall(vars_regex, command.command)
            for typ, val in re_groups:

                sub_str_original = f"<<{typ}:{val}>>"

                if typ == "executable":
                    if env is None:
                        env = self.get_environment()
                    exe = env.executables.get(val)
                    sub_str_new = exe.instances[0].command  # TODO: ...

                elif typ == "parameter":
                    param = self.element.get(f"inputs.{val}")
                    sub_str_new = str(param)  # TODO: custom formatting...

                elif typ == "script":
                    script_name = val
                    sub_str_new = '"' + str(self.element.dir_path / script_name) + '"'
                    scripts.append(script_name)

                elif typ == "file":
                    sub_str_new = self.app.command_files.get(val).value()

                command_resolved = command_resolved.replace(sub_str_original, sub_str_new)

            resolved_commands.append(command_resolved)

        # generate scripts:
        for script in scripts:
            script_path = self.element.dir_path / script
            snippet_path = self.app.scripts.get(script)
            with snippet_path.open("rt") as fp:
                script_body = fp.readlines()

            main_func_name = script.strip(".py")  # TODO: don't assume this

            script_lns = script_body
            script_lns += [
                "\n\n",
                'if __name__ == "__main__":\n',
                "    import zarr\n",
            ]

            if self.input_file_generator:
                input_file = self.input_file_generator.input_file
                invoc_args = f"path=Path('./{input_file.value()}'), **params"
                input_zarr_groups = {
                    k.typ: self.element.data_index[f"inputs.{k.typ}"]
                    for k in self.input_file_generator.inputs
                }
                script_lns += [
                    f"    from hpcflow.sdk.core.zarr_io import zarr_decode\n\n",
                    f"    params = {{}}\n",
                    f"    param_data = Path('../../../parameter_data')\n",
                    f"    for param_group_idx in {list(input_zarr_groups.values())!r}:\n",
                ]
                for k in input_zarr_groups:
                    script_lns += [
                        f"        grp_i = zarr.open(param_data / str(param_group_idx), mode='r')\n",
                        f"        params[{k!r}] = zarr_decode(grp_i)\n",
                    ]

                script_lns += [
                    f"\n    {main_func_name}({invoc_args})\n\n",
                ]

            elif self.output_parser:
                out_name = self.output_parser.output.typ
                out_files = {k.label: k.value() for k in self.output_parser.output_files}
                invoc_args = ", ".join(f"{k}={v!r}" for k, v in out_files.items())
                output_zarr_group = self.element.data_index[f"outputs.{out_name}"]

                script_lns += [
                    f"    from hpcflow.sdk.core.zarr_io import zarr_encode\n\n",
                    f"    {out_name} = {main_func_name}({invoc_args})\n\n",
                ]

                script_lns += [
                    f"    param_data = Path('../../../parameter_data')\n",
                    f"    output_group = zarr.open(param_data / \"{str(output_zarr_group)}\", mode='r+')\n",
                    f"    zarr_encode({out_name}, output_group)\n",
                ]

            with script_path.open("wt", newline="") as fp:
                fp.write("".join(script_lns))

        for command in resolved_commands:
            proc_i = subprocess.run(
                command,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=self.element.dir_path,
            )
            stdout = proc_i.stdout.decode()
            stderr = proc_i.stderr.decode()
            if stdout:
                print(stdout)
            if stderr:
                print(stderr)


class ActionScope(JSONLike):
    """Class to represent the identification of a subset of task schema actions by a
    filtering process.
    """

    _child_objects = (
        ChildObjectSpec(
            name="typ",
            json_like_name="type",
            class_name="ActionScopeType",
            is_enum=True,
        ),
    )

    def __init__(self, typ: Union[ActionScopeType, str], **kwargs):

        if isinstance(typ, str):
            typ = getattr(self.app.ActionScopeType, typ.upper())

        self.typ = typ
        self.kwargs = {k: v for k, v in kwargs.items() if v is not None}

        bad_keys = set(kwargs.keys()) - ACTION_SCOPE_ALLOWED_KWARGS[self.typ.name]
        if bad_keys:
            raise TypeError(
                f"The following keyword arguments are unknown for ActionScopeType "
                f"{self.typ.name}: {bad_keys}."
            )

    def __repr__(self):
        kwargs_str = ""
        if self.kwargs:
            kwargs_str = ", ".join(f"{k}={v!r}" for k, v in self.kwargs.items())
        return f"{self.__class__.__name__}.{self.typ.name.lower()}({kwargs_str})"

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self.typ is other.typ and self.kwargs == other.kwargs:
            return True
        return False

    @classmethod
    def _parse_from_string(cls, string):
        typ_str, kwargs_str = re.search(ACTION_SCOPE_REGEX, string).groups()
        kwargs = {}
        if kwargs_str:
            for i in kwargs_str.split(","):
                name, val = i.split("=")
                kwargs[name.strip()] = val.strip()
        return {"type": typ_str, **kwargs}

    def to_string(self):
        kwargs_str = ""
        if self.kwargs:
            kwargs_str = "[" + ", ".join(f"{k}={v}" for k, v in self.kwargs.items()) + "]"
        return f"{self.typ.name.lower()}{kwargs_str}"

    @classmethod
    def from_json_like(cls, json_like, shared_data=None):
        if isinstance(json_like, str):
            json_like = cls._parse_from_string(json_like)
        else:
            typ = json_like.pop("type")
            json_like = {"type": typ, **json_like.pop("kwargs", {})}
        return super().from_json_like(json_like, shared_data)

    @classmethod
    def any(cls):
        return cls(typ=ActionScopeType.ANY)

    @classmethod
    def main(cls):
        return cls(typ=ActionScopeType.MAIN)

    @classmethod
    def processing(cls):
        return cls(typ=ActionScopeType.PROCESSING)

    @classmethod
    def input_file_generator(cls, file=None):
        return cls(typ=ActionScopeType.INPUT_FILE_GENERATOR, file=file)

    @classmethod
    def output_file_parser(cls, output=None):
        return cls(typ=ActionScopeType.OUTPUT_FILE_PARSER, output=output)


@dataclass
class ActionEnvironment(JSONLike):

    _app_attr = "app"

    _child_objects = (
        ChildObjectSpec(
            name="scope",
            class_name="ActionScope",
        ),
        ChildObjectSpec(
            name="environment",
            class_name="Environment",
            shared_data_name="environments",
            shared_data_primary_key="name",
        ),
    )

    environment: Environment
    scope: Optional[ActionScope] = None

    def __post_init__(self):
        if self.scope is None:
            self.scope = self.app.ActionScope.any()


@dataclass
class ActionRule(JSONLike):
    """Class to represent a rule/condition that must be True if an action is to be
    included."""

    _app_attr = "app"

    _child_objects = (ChildObjectSpec(name="rule", class_obj=Rule),)

    check_exists: Optional[str] = None
    check_missing: Optional[str] = None
    rule: Optional[Rule] = None

    def __post_init__(self):
        if (
            self.check_exists is not None
            and self.check_missing is not None
            and self.rule is not None
        ) or (
            self.check_exists is None and self.check_missing is None and self.rule is None
        ):
            raise ValueError(
                "Specify exactly one of `check_exists`, `check_missing` and `rule`."
            )

    def __repr__(self):

        out = f"{self.__class__.__name__}("
        if self.check_exists:
            out += f"check_exists={self.check_exists!r}"
        elif self.check_missing:
            out += f"check_missing={self.check_missing!r}"
        else:
            out += f"rule={self.rule}"
        out += ")"
        return out


class Action(JSONLike):
    """"""

    _app_attr = "app"
    _child_objects = (
        ChildObjectSpec(
            name="commands",
            class_name="Command",
            is_multiple=True,
        ),
        ChildObjectSpec(
            name="input_file_generators",
            is_multiple=True,
            class_name="InputFileGenerator",
            dict_key_attr="input_file",
        ),
        ChildObjectSpec(
            name="output_file_parsers",
            is_multiple=True,
            class_name="OutputFileParser",
            dict_key_attr="output",
        ),
        ChildObjectSpec(
            name="input_files",
            is_multiple=True,
            class_name="FileSpec",
            shared_data_name="command_files",
        ),
        ChildObjectSpec(
            name="output_files",
            is_multiple=True,
            class_name="FileSpec",
            shared_data_name="command_files",
        ),
        ChildObjectSpec(
            name="environments",
            class_name="ActionEnvironment",
            is_multiple=True,
        ),
        ChildObjectSpec(
            name="rules",
            class_name="ActionRule",
            is_multiple=True,
        ),
    )

    def __init__(
        self,
        commands: List[Command],
        environments: List[ActionEnvironment],
        input_file_generators: Optional[List[InputFileGenerator]] = None,
        output_file_parsers: Optional[List[OutputFileParser]] = None,
        input_files: Optional[List[FileSpec]] = None,
        output_files: Optional[List[FileSpec]] = None,
        rules: Optional[List[ActionRule]] = None,
    ):
        self.commands = commands
        self.environments = environments
        self.input_file_generators = input_file_generators or []
        self.output_file_parsers = output_file_parsers or []
        self.input_files = self._resolve_input_files(input_files or [])
        self.output_files = self._resolve_output_files(output_files or [])
        self.rules = rules or []

        self._from_expand = False  # assigned on creation of new Action by `expand`

    def _resolve_input_files(self, input_files):
        in_files = input_files
        for i in self.input_file_generators:
            if i.input_file not in in_files:
                in_files.append(i.input_file)
        return in_files

    def _resolve_output_files(self, output_files):
        out_files = output_files
        for i in self.output_file_parsers:
            for j in i.output_files:
                if j not in out_files:
                    out_files.append(j)
        return out_files

    def __repr__(self) -> str:
        IFGs = {
            i.input_file.label: [j.typ for j in i.inputs]
            for i in self.input_file_generators
        }
        OFPs = {
            i.output.typ: [j.label for j in i.output_files]
            for i in self.output_file_parsers
        }

        out = []
        if self.commands:
            out.append(f"commands={self.commands!r}")
        if self.environments:
            out.append(f"environments={self.environments!r}")
        if IFGs:
            out.append(f"input_file_generators={IFGs!r}")
        if OFPs:
            out.append(f"output_file_parsers={OFPs!r}")
        if self.rules:
            out.append(f"rules={self.rules!r}")

        return f"{self.__class__.__name__}({', '.join(out)})"

    def __eq__(self, other):
        if type(other) is not self.__class__:
            return False
        if (
            self.commands == other.commands
            and self.environments == other.environments
            and self.input_file_generators == other.input_file_generators
            and self.output_file_parsers == other.output_file_parsers
            and self.rules == other.rules
        ):
            return True
        return False

    @classmethod
    def _json_like_constructor(cls, json_like):
        """Invoked by `JSONLike.from_json_like` instead of `__init__`."""
        _from_expand = json_like.pop("_from_expand", None)
        obj = cls(**json_like)
        obj._from_expand = _from_expand
        return obj

    def get_parameter_dependence(self, parameter: SchemaParameter):
        """Find if/where a given parameter is used by the action."""
        writer_files = [
            i.input_file
            for i in self.input_file_generators
            if parameter.parameter in i.inputs
        ]  # names of input files whose generation requires this parameter
        commands = []  # TODO: indices of commands in which this parameter appears
        out = {"input_file_writers": writer_files, "commands": commands}
        return out

    def get_resolved_action_env(
        self,
        relevant_scopes: Tuple[ActionScopeType],
        input_file_generator: InputFileGenerator = None,
        output_file_parser: OutputFileParser = None,
        commands: List[Command] = None,
    ):
        possible = [i for i in self.environments if i.scope.typ in relevant_scopes]
        if not possible:
            if input_file_generator:
                msg = f"input file generator {input_file_generator.input_file.label!r}"
            elif output_file_parser:
                msg = f"output file parser {output_file_parser.output.typ!r}"
            else:
                msg = f"commands {commands!r}"
            raise MissingCompatibleActionEnvironment(
                f"No compatible environment is specified for the {msg}."
            )

        # sort by scope type specificity:
        possible_srt = sorted(possible, key=lambda i: i.scope.typ.value, reverse=True)
        return possible_srt[0]

    def get_input_file_generator_action_env(
        self, input_file_generator: InputFileGenerator
    ):
        return self.get_resolved_action_env(
            relevant_scopes=(
                ActionScopeType.ANY,
                ActionScopeType.PROCESSING,
                ActionScopeType.INPUT_FILE_GENERATOR,
            ),
            input_file_generator=input_file_generator,
        )

    def get_output_file_parser_action_env(self, output_file_parser: OutputFileParser):
        return self.get_resolved_action_env(
            relevant_scopes=(
                ActionScopeType.ANY,
                ActionScopeType.PROCESSING,
                ActionScopeType.OUTPUT_FILE_PARSER,
            ),
            output_file_parser=output_file_parser,
        )

    def get_commands_action_env(self):
        return self.get_resolved_action_env(
            relevant_scopes=(ActionScopeType.ANY, ActionScopeType.MAIN),
            commands=self.commands,
        )

    def expand(self):
        if self._from_expand:
            # already expanded
            return [self]

        else:

            # run main if:
            #   - one or more output files are not passed
            # run IFG if:
            #   - one or more output files are not passed
            #   - AND input file is not passed
            # always run OPs, for now

            out_file_rules = [
                self.app.ActionRule(check_missing=f"output_files.{j.label}")
                for i in self.output_file_parsers
                for j in i.output_files
            ]

            main_rules = self.rules + out_file_rules

            # note we keep the IFG/OPs in the new actions, so we can check the parameters
            # used/produced.

            inp_files = []
            inp_acts = []
            for IFG_i in self.input_file_generators:
                script = "script-name"  # TODO
                act_i = self.app.Action(
                    commands=[
                        self.app.Command(f"<<executable:python>> <<script:{script}>>")
                    ],
                    input_file_generators=[IFG_i],
                    environments=[self.get_input_file_generator_action_env(IFG_i)],
                    rules=main_rules + [IFG_i.get_action_rule()],
                )
                inp_files.append(IFG_i.input_file)
                act_i._from_expand = True
                inp_acts.append(act_i)

            out_files = []
            out_acts = []
            for OP_i in self.output_file_parsers:
                script = "script-name"  # TODO
                act_i = self.app.Action(
                    commands=[
                        self.app.Command(f"<<executable:python>> <<script:{script}>>")
                    ],
                    output_file_parsers=[OP_i],
                    environments=[self.get_output_file_parser_action_env(OP_i)],
                    rules=list(self.rules),
                )
                out_files.extend(OP_i.output_files)
                act_i._from_expand = True
                out_acts.append(act_i)

            main_act = self.app.Action(
                commands=self.commands,
                environments=[self.get_commands_action_env()],
                rules=main_rules,
                input_files=inp_files,
                output_files=out_files,
            )
            main_act._from_expand = True

            cmd_acts = inp_acts + [main_act] + out_acts

            return cmd_acts

    def get_command_input_types(self) -> Tuple[str]:
        """Get parameter types from commands."""
        params = []
        # note: we use "parameter" rather than "input", because it could be a schema input
        # or schema output.
        vars_regex = r"\<\<parameter:(.*?)\>\>"
        for command in self.commands:
            for val in re.findall(vars_regex, command.command):
                params.append(val)
            # TODO: consider stdin?
        return tuple(set(params))

    def get_command_output_types(self) -> Tuple[str]:
        """Get parameter types from command stdout and stderr arguments."""
        params = []
        for command in self.commands:
            out_params = command.get_output_types()
            if out_params["stdout"]:
                params.append(out_params["stdout"])
            if out_params["stderr"]:
                params.append(out_params["stderr"])

        return tuple(set(params))

    def get_input_types(self) -> Tuple[str]:
        """Get the input types that are consumed by commands and input file generators of
        this action."""
        params = list(self.get_command_input_types())
        for i in self.input_file_generators:
            params.extend([j.typ for j in i.inputs])
        return tuple(set(params))

    def get_output_types(self) -> Tuple[str]:
        """Get the output types that are produced by command standard outputs and errors,
        and by output file parsers of this action."""
        params = list(self.get_command_output_types())
        for i in self.output_file_parsers:
            params.append(i.output.typ)
        return tuple(params)

    def get_input_file_labels(self):
        return tuple(i.label for i in self.input_files)

    def get_output_file_labels(self):
        return tuple(i.label for i in self.output_files)

    def generate_data_index(
        self, act_idx, EAR_idx, schema_data_idx, all_data_idx, workflow, param_source
    ):
        """Generate the data index for this action of an element iteration whose overall
        data index is passed."""

        # output keys must be processed first for this to work, since when processing an
        # output key, we may need to update the index of an output in a previous action's
        # data index, which could affect the data index in an input of this action.
        keys = [f"outputs.{i}" for i in self.get_output_types()]
        keys += [f"inputs.{i}" for i in self.get_input_types()]
        for i in self.input_files:
            keys.append(f"input_files.{i.label}")
        for i in self.output_files:
            keys.append(f"output_files.{i.label}")

        # keep all resources data:
        sub_data_idx = {k: v for k, v in schema_data_idx.items() if "resources" in k}
        param_src_update = []
        for key in keys:
            sub_param_idx = {}
            if (
                key.startswith("input_files")
                or key.startswith("output_files")
                or key.startswith("inputs")
            ):
                # look for an index in previous data indices (where for inputs we look
                # for *output* parameters of the same name):
                k_idx = None
                for prev_data_idx in all_data_idx.values():

                    if key.startswith("inputs"):
                        k_param = key.split("inputs.")[1]
                        k_out = f"outputs.{k_param}"
                        if k_out in prev_data_idx:
                            k_idx = prev_data_idx[k_out]

                    else:
                        if key in prev_data_idx:
                            k_idx = prev_data_idx[key]

                if k_idx is None:
                    # otherwise take from the schema_data_idx:
                    if key in schema_data_idx:
                        k_idx = schema_data_idx[key]
                        # add any associated sub-parameters:
                        for k, v in schema_data_idx.items():
                            if k.startswith(f"{key}."):  # sub-parameter (note dot)
                                sub_param_idx[k] = v
                    else:
                        # otherwise we need to allocate a new parameter datum:
                        # (for input/output_files keys)
                        k_idx = workflow._add_unset_parameter_data(param_source)

            else:
                # outputs
                k_idx = None
                for (act_idx_i, EAR_idx_i), prev_data_idx in all_data_idx.items():
                    if key in prev_data_idx:

                        k_idx = prev_data_idx[key]

                        # allocate a new parameter datum for this intermediate output:
                        param_source_i = copy.deepcopy(param_source)
                        param_source_i["action_idx"] = act_idx_i
                        param_source_i["EAR_idx"] = EAR_idx_i
                        new_k_idx = workflow._add_unset_parameter_data(param_source_i)
                        prev_data_idx[key] = new_k_idx
                if k_idx is None:
                    # otherwise take from the schema_data_idx:
                    k_idx = schema_data_idx[key]

                # can now set the EAR/act idx in the associated parameter source
                param_src_update.append(k_idx)

            sub_data_idx[key] = k_idx
            sub_data_idx.update(sub_param_idx)

        all_data_idx[(act_idx, EAR_idx)] = sub_data_idx

        return param_src_update

    def get_possible_scopes(self) -> Tuple[ActionScope]:
        """Get the action scopes that are inclusive of this action, ordered by decreasing
        specificity."""

        scope = self.get_precise_scope()

        if self.input_file_generators:
            scopes = (
                scope,
                self.app.ActionScope.input_file_generator(),
                self.app.ActionScope.processing(),
                self.app.ActionScope.any(),
            )
        elif self.output_file_parsers:
            scopes = (
                scope,
                self.app.ActionScope.output_file_parser(),
                self.app.ActionScope.processing(),
                self.app.ActionScope.any(),
            )
        else:
            scopes = (scope, self.app.ActionScope.any())

        return scopes

    def get_precise_scope(self) -> ActionScope:
        if not self._from_expand:
            raise RuntimeError(
                "Precise scope cannot be unambiguously defined until the Action has been "
                "expanded."
            )

        if self.input_file_generators:
            return self.app.ActionScope.input_file_generator(
                file=self.input_file_generators[0].input_file.label
            )
        elif self.output_file_parsers:
            return self.app.ActionScope.output_file_parser(
                output=self.output_file_parsers[0].output.typ
            )
        else:
            return self.app.ActionScope.main()

    def is_input_type_required(self, typ: str, provided_files: List[FileSpec]) -> bool:

        # typ is required if is appears in any command:
        if typ in self.get_command_input_types():
            return True

        # typ is required if used in any input file generators and input file is not
        # provided:
        for IFG in self.input_file_generators:
            if typ in (i.typ for i in IFG.inputs):
                if IFG.input_file not in provided_files:
                    return True



from __future__ import annotations

import copy
import functools
import json
import logging
import socket
import uuid
from dataclasses import dataclass, field
from hashlib import new
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from fsspec.registry import known_implementations as fsspec_protocols
from platformdirs import user_data_dir
from valida.schema import Schema

from hpcflow.sdk.core.validation import get_schema
from hpcflow.sdk.log import AppLog

from .callbacks import (
    callback_bool,
    callback_file_paths,
    set_callback_file_paths,
    check_load_data_files,
)
from .config_file import ConfigFile
from .errors import (
    ConfigChangeInvalidJSONError,
    ConfigChangePopIndexError,
    ConfigChangeTypeInvalidError,
    ConfigChangeValidationError,
    ConfigItemAlreadyUnsetError,
    ConfigItemCallbackError,
    ConfigNonConfigurableError,
    ConfigUnknownItemError,
    ConfigUnknownOverrideError,
    ConfigValidationError,
)

logger = logging.getLogger(__name__)

DEFAULT_CONFIG_FILE = {
    "configs": {
        "default": {
            "invocation": {"environment_setup": None, "match": {}},
            "config": {
                "machine": socket.gethostname(),
                "telemetry": True,
                "log_file_path": "logs/app.log",
                "environment_sources": [],
                "task_schema_sources": [],
                "command_file_sources": [],
                "parameter_sources": [],
            },
        }
    }
}


@dataclass
class ConfigOptions:
    """Application-level options for configuration"""

    default_directory: Union[Path, str]
    directory_env_var: str
    sentry_DSN: str
    sentry_traces_sample_rate: float
    sentry_env: str
    default_config: Optional[Dict] = field(
        default_factory=lambda: copy.deepcopy(DEFAULT_CONFIG_FILE)
    )
    extra_schemas: Optional[List[Schema]] = field(default_factory=lambda: [])


class Config:
    """Application configuration as defined in one or more config files.

    Methods
    -------
    to_string : get a formatted string
    get_configurable : list all configurable items
    get : get a configuration item
    set_value : set a configuration item
    unset_value : unset a configuration item
    append_value : append a value to a list configuration item
    prepend_value : prepend a value to a list configuration item
    pop_value : pop a value from a list configuration item

    Notes
    -----
    On modifying/setting existing values, modifications are not automatically copied
    to the configuration file; use `save()` to save to the file. Items in `overrides`
    are not saved into the file.

    """

    def __init__(
        self,
        app,
        options: ConfigOptions,
        config_dir,
        config_invocation_key,
        logger,
        uid=None,
        callbacks=None,
        **overrides,
    ):

        self._app = app
        self._options = options
        self._overrides = overrides
        self._logger = logger

        # Callbacks are run on get:
        self._get_callbacks = {
            "task_schema_sources": (callback_file_paths,),
            "environment_sources": (callback_file_paths,),
            "parameter_sources": (callback_file_paths,),
            "command_file_sources": (callback_file_paths,),
            "log_file_path": (callback_file_paths,),
            "telemetry": (callback_bool,),
            **(callbacks or {}),
        }

        # Set callbacks are run on set:
        self._set_callbacks = {
            "task_schema_sources": (set_callback_file_paths, check_load_data_files),
            "environment_sources": (set_callback_file_paths, check_load_data_files),
            "parameter_sources": (set_callback_file_paths, check_load_data_files),
            "command_file_sources": (set_callback_file_paths, check_load_data_files),
            "log_file_path": (set_callback_file_paths,),
        }

        cfg_schemas, cfg_keys = self._init_schemas()
        self._schemas = cfg_schemas

        self._file = ConfigFile(self, config_dir, config_invocation_key)

        self._configurable_keys = cfg_keys
        self._modified_keys = {}
        self._unset_keys = []

        for name in overrides:
            if name not in self._configurable_keys:
                raise ConfigUnknownOverrideError(name=name)

        host_uid, host_uid_file_path = self._get_user_id()

        self._meta_data = {
            "config_directory": self._file.directory,
            "config_file_name": self._file.path.name,
            "config_file_path": self._file.path,
            "config_file_contents": self._file.contents,
            "config_invocation_key": self._file.invoc_key,
            "config_schemas": cfg_schemas,
            "invoking_user_id": uid or host_uid,
            "host_user_id": host_uid,
            "host_user_id_file_path": host_uid_file_path,
        }

        self._validate()

    def __str__(self):
        return self.to_string(exclude=["config_file_contents"])

    def __dir__(self):
        return super().__dir__() + self._all_keys

    def __getattr__(self, name):
        return self.get(name)

    def __setattr__(self, name, value):
        if (
            "_configurable_keys" in self.__dict__
            and name in self.__dict__["_configurable_keys"]
        ):
            self.set(name, value)
        else:
            super().__setattr__(name, value)

    def _validate(self, data=None, raise_with_metadata=True):
        """Validate configuration items of the loaded invocation."""

        self._logger.debug("Validating configuration...")
        if data is None:
            validated_data = self.get_all(include_overrides=True)
        else:
            validated_data = data

        for cfg_schema in self._schemas:
            cfg_validated = cfg_schema.validate(validated_data)
            if not cfg_validated.is_valid:
                meta_data = None
                if raise_with_metadata:
                    meta_data = self.to_string(
                        exclude=["config_file_contents", "config_schemas"], just_meta=True
                    )
                raise ConfigValidationError(
                    message=cfg_validated.get_failures_string(),
                    meta_data=meta_data,
                )
            validated_data = cfg_validated.cast_data

        self._logger.debug("Configuration is valid.")
        return validated_data

    def to_string(self, exclude: Optional[List] = None, just_meta=False):
        """Format the instance in a string, optionally exclude some keys.

        Parameters
        ----------
        exclude
            List of keys to exclude. Optional.
        just_meta
            If True, just return a str of the meta-data. This is useful to show during
            initialisation, in the case where the configuration is otherwise invalid.

        """
        exclude = exclude or []
        lines = []
        blocks = {"meta-data": self._meta_data}
        if not just_meta:
            blocks.update({"configuration": self.get_all(as_str=True)})
        for title, dat in blocks.items():
            lines.append(f"{title}:")
            for key, val in dat.items():
                if key in exclude:
                    continue
                if isinstance(val, list):
                    if val:
                        val = "\n    " + "\n    ".join(str(i) for i in val)
                    else:
                        val = "[]"
                lines.append(f"  {key}: {val}")
        return "\n".join(lines)

    def _resolve_path(self, path):
        """Resolve a file path, but leave fsspec protocols alone."""
        if not any(str(path).startswith(i + ":") for i in fsspec_protocols):
            path = Path(path)
            path = path.expanduser()
            if not path.is_absolute():
                path = self._meta_data["config_directory"].joinpath(path)
        else:
            self._logger.debug(
                f"Not resolving path {path!r} because it looks like an `fsspec` URL."
            )

        return path

    def register_config_get_callback(self, name):
        """Decorator to register a function as a configuration callback for a specified
        configuration item name, to be invoked on `get` of the item."""

        def decorator(func):
            if name in self._get_callbacks:
                self._get_callbacks[name] = tuple(
                    list(self._get_callbacks[name]) + [func]
                )
            else:
                self._get_callbacks[name] = (func,)

            @functools.wraps(func)
            def wrap(value):
                return func(value)

            return wrap

        return decorator

    def register_config_set_callback(self, name):
        """Decorator to register a function as a configuration callback for a specified
        configuration item name, to be invoked on `set` of the item."""

        def decorator(func):
            if name in self._set_callbacks:
                self._set_callbacks[name] = tuple(
                    list(self._set_callbacks[name]) + [func]
                )
            else:
                self._set_callbacks[name] = (func,)

            @functools.wraps(func)
            def wrap(value):
                return func(value)

            return wrap

        return decorator

    @property
    def _all_keys(self):
        return self._configurable_keys + list(self._meta_data.keys())

    def get_all(self, include_overrides=True, as_str=False):
        """Get all configurable items."""
        items = {}
        for key in self._configurable_keys:
            if key in self._unset_keys:
                continue
            else:
                try:
                    val = self.get(
                        name=key,
                        include_overrides=include_overrides,
                        raise_on_missing=True,
                        as_str=as_str,
                    )
                except ValueError:
                    continue
                items.update({key: val})
        return items

    def _get_callback_value(self, name, value):
        if name in self._get_callbacks and value is not None:
            for cb in self._get_callbacks.get(name, []):
                self._logger.debug(
                    f"Invoking `config.get` callback ({cb.__name__!r}) for item {name!r}={value!r}"
                )
                value = cb(self, value)
        return value

    def get(
        self,
        name,
        include_overrides=True,
        raise_on_missing=False,
        as_str=False,
        callback=True,
        default_value=None,
    ):
        """Get a configuration item.

        Notes
        -----

        """
        if name not in self._all_keys:
            raise ConfigUnknownItemError(name=name)

        elif name in self._meta_data:
            val = self._meta_data[name]

        elif include_overrides and name in self._overrides:
            val = self._overrides[name]

        elif name in self._unset_keys:
            if raise_on_missing:
                raise ValueError("Not set.")
            val = None
            if default_value:
                val = default_value

        elif name in self._modified_keys:
            val = self._modified_keys[name]

        elif name in self._configurable_keys:
            val = self._file.get_config_item(
                name, raise_on_missing, default_value=default_value
            )

        if callback:
            try:
                val = self._get_callback_value(name, val)
            except Exception as err:
                raise ConfigItemCallbackError(name, err)

        if as_str:
            if isinstance(val, (list, tuple, set)):
                val = [str(i) for i in val]
            else:
                val = str(val)

        return val

    def _parse_JSON(self, name, value):
        try:
            value = json.loads(value)
        except json.decoder.JSONDecodeError as err:
            raise ConfigChangeInvalidJSONError(name=name, json_str=value, err=err)
        return value

    def set(self, name, value, is_json=False, callback=True):

        self._logger.debug(f"Attempting to set config item {name!r} to {value!r}.")

        if name not in self._configurable_keys:
            raise ConfigNonConfigurableError(name=name)
        else:
            if is_json:
                value = self._parse_JSON(name, value)
            current_val = self.get(name)
            callback_val = self._get_callback_value(name, value)
            file_val_raw = self._file.get_config_item(name)
            file_val = self._get_callback_value(name, file_val_raw)

            if callback_val != current_val:

                was_in_modified = False
                was_in_unset = False
                prev_modified_val = None
                modified_updated = False

                if name in self._modified_keys:
                    was_in_modified = True
                    prev_modified_val = self._modified_keys[name]

                if name in self._unset_keys:
                    was_in_unset = True
                    idx = self._unset_keys.index(name)
                    self._unset_keys.pop(idx)

                if callback_val != file_val:
                    self._modified_keys[name] = value
                    modified_updated = True

                try:
                    self._validate()

                    if callback:
                        for cb in self._set_callbacks.get(name, []):
                            self._logger.debug(
                                f"Invoking `config.set` callback for item {name!r}: {cb.__name__!r}"
                            )
                            cb(self, callback_val)

                except ConfigValidationError as err:

                    # revert:
                    if modified_updated:
                        if was_in_modified:
                            self._modified_keys[name] = prev_modified_val
                        else:
                            del self._modified_keys[name]
                    if was_in_unset:
                        self._unset_keys.append(name)

                    raise ConfigChangeValidationError(name, validation_err=err) from None

                self._logger.debug(
                    f"Successfully set config item {name!r} to {callback_val!r}."
                )
            else:
                print(f"value is already: {callback_val!r}")

    def unset(self, name):

        if name not in self._configurable_keys:
            raise ConfigNonConfigurableError(name=name)
        if name in self._unset_keys or not self._file.is_item_set(name):
            raise ConfigItemAlreadyUnsetError(name=name)

        self._unset_keys.append(name)
        try:
            self._validate()
        except ConfigValidationError as err:
            self._unset_keys.pop()
            raise ConfigChangeValidationError(name, validation_err=err) from None

    def append(self, name, value, is_json=False):
        existing = self.get(name, callback=False)
        if is_json:
            value = self._parse_JSON(name, value)
        try:
            new = existing + [value]
        except TypeError:
            raise ConfigChangeTypeInvalidError(name, typ=type(existing)) from None

        self.set(name, new)

    def prepend(self, name, value, is_json=False):
        existing = self.get(name, callback=False)
        if is_json:
            value = self._parse_JSON(name, value)
        try:
            new = [value] + existing
        except TypeError:
            raise ConfigChangeTypeInvalidError(name, typ=type(existing)) from None

        self.set(name, new)

    def pop(self, name, index):
        existing = self.get(name, callback=False)
        try:
            new = copy.deepcopy(existing)
            new.pop(index)
        except AttributeError:
            raise ConfigChangeTypeInvalidError(name, typ=type(existing)) from None
        except IndexError:
            raise ConfigChangePopIndexError(
                name, length=len(existing), index=index
            ) from None

        self.set(name, new)

    def save(self):
        """Save any modified/unset configuration items into the file."""
        if not self._modified_keys and not self._unset_keys:
            print("No modifications to save!")
        else:
            self._file.save()

    def get_configurable(self):
        """Get a list of all configurable keys."""
        return self._configurable_keys

    def _get_user_id(self):
        """Retrieve (and set if non-existent) a unique user ID that is independent of the
        config directory."""

        uid_file_dir = Path(user_data_dir(appname=self._app.name))
        uid_file_path = uid_file_dir.joinpath("user_id.txt")
        if not uid_file_path.exists():
            uid_file_dir.mkdir(exist_ok=True, parents=True)
            uid = str(uuid.uuid4())
            with uid_file_path.open("wt") as fh:
                fh.write(uid)
        else:
            with uid_file_path.open("rt") as fh:
                uid = fh.read().strip()

        # Generate sub-dir for this hostname (used by helper process):
        hostname_dir = uid_file_dir.joinpath(socket.gethostname())
        hostname_dir.mkdir(exist_ok=True)

        return uid, uid_file_path

    def _init_schemas(self):
        # Get allowed configurable keys from config schemas:
        cfg_schemas = [get_schema("config_schema.yaml")] + self._options.extra_schemas
        cfg_keys = []
        for cfg_schema in cfg_schemas:
            for rule in cfg_schema.rules:
                if not rule.path and rule.condition.callable.name == "allowed_keys":
                    cfg_keys.extend(rule.condition.callable.args)

        return (cfg_schemas, cfg_keys)



from __future__ import annotations
from contextlib import contextmanager
import copy
from dataclasses import dataclass, field
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Tuple, Type, Union
from warnings import warn

import numpy as np

from hpcflow.sdk.core.actions import EAR_ID
from hpcflow.sdk.submission.jobscript import (
    generate_EAR_resource_map,
    group_resource_map_into_jobscripts,
    jobscripts_to_list,
    merge_jobscripts_across_tasks,
    resolve_jobscript_dependencies,
)
from hpcflow.sdk.typing import PathLike

from .json_like import ChildObjectSpec, JSONLike
from .parameters import InputSource
from .task import ElementSet, Task
from .utils import (
    read_JSON_file,
    read_JSON_string,
    read_YAML,
    read_YAML_file,
    replace_items,
)
from .errors import (
    InvalidInputSourceTaskReference,
    LoopAlreadyExistsError,
    SubmissionFailure,
    WorkflowBatchUpdateFailedError,
    WorkflowNotFoundError,
    WorkflowSubmissionFailure,
)
from hpcflow.sdk.persistence import (
    store_cls_from_path,
    store_cls_from_str,
    temporary_workflow_rename,
    DEFAULT_STORE_FORMAT,
)

DEFAULT_TEMPLATE_FORMAT = "yaml"
ALL_TEMPLATE_FORMATS = ("yaml", "json")


class _DummyPersistentWorkflow:
    """An object to pass to ResourceSpec.make_persistent that pretends to be a
    Workflow object, so we can pretend to make template-level inputs/resources
    persistent before the workflow exists."""

    def __init__(self):
        self._parameters = []
        self._sources = []

    def _add_parameter_data(self, data, source: Dict) -> int:
        self._parameters.append(data)
        self._sources.append(source)
        data_ref = len(self._parameters) - 1
        return data_ref

    def make_persistent(self, workflow: Workflow):
        for dat_i, source_i in zip(self._parameters, self._sources):
            workflow._add_parameter_data(dat_i, source_i)


@dataclass
class WorkflowTemplate(JSONLike):
    """Class to represent initial parametrisation of a {app_name} workflow, with limited
    validation logic."""

    _app_attr = "app"

    _child_objects = (
        ChildObjectSpec(
            name="tasks",
            class_name="Task",
            is_multiple=True,
            parent_ref="workflow_template",
        ),
        ChildObjectSpec(
            name="loops",
            class_name="Loop",
            is_multiple=True,
            parent_ref="_workflow_template",
        ),
        ChildObjectSpec(
            name="resources",
            class_name="ResourceList",
            parent_ref="_workflow_template",
        ),
    )

    name: str
    tasks: Optional[List[Task]] = field(default_factory=lambda: [])
    loops: Optional[List[Loop]] = field(default_factory=lambda: [])
    workflow: Optional[Workflow] = None
    resources: Optional[Dict[str, Dict]] = None

    def __post_init__(self):

        if isinstance(self.resources, dict):
            self.resources = self.app.ResourceList.from_json_like(self.resources)
        elif isinstance(self.resources, list):
            self.resources = self.app.ResourceList(self.resources)
        elif not self.resources:
            self.resources = self.app.ResourceList([self.app.ResourceSpec()])

        self._set_parent_refs()

    @classmethod
    def _from_data(cls, data: Dict) -> WorkflowTemplate:
        # use element_sets if not already:
        for task_idx, task_dat in enumerate(data["tasks"]):
            if "element_sets" not in task_dat:
                # add a single element set:
                elem_set = {}
                for chd_obj in ElementSet._child_objects:
                    if chd_obj.name in task_dat:
                        elem_set[chd_obj.name] = task_dat.pop(chd_obj.name)
                data["tasks"][task_idx]["element_sets"] = [elem_set]

        # extract out any template components:
        params_dat = data.pop("parameters", [])
        if params_dat:
            parameters = cls.app.ParametersList.from_json_like(
                params_dat, shared_data=cls.app.template_components
            )
            cls.app.parameters.add_objects(parameters, skip_duplicates=True)

        cmd_files_dat = data.pop("command_files", [])
        if cmd_files_dat:
            cmd_files = cls.app.CommandFilesList.from_json_like(
                cmd_files_dat, shared_data=cls.app.template_components
            )
            cls.app.command_files.add_objects(cmd_files, skip_duplicates=True)

        envs_dat = data.pop("environments", [])
        if envs_dat:
            envs = cls.app.EnvironmentsList.from_json_like(
                envs_dat, shared_data=cls.app.template_components
            )
            cls.app.envs.add_objects(envs, skip_duplicates=True)

        ts_dat = data.pop("task_schemas", [])
        if ts_dat:
            task_schemas = cls.app.TaskSchemasList.from_json_like(
                ts_dat, shared_data=cls.app.template_components
            )
            cls.app.task_schemas.add_objects(task_schemas, skip_duplicates=True)

        return cls.from_json_like(data, shared_data=cls.app.template_components)

    @classmethod
    def from_YAML_string(cls, string: str) -> WorkflowTemplate:
        """Load from a YAML string.

        Parameters
        ----------
        string
            The YAML string containing the workflow template parametrisation.

        """
        return cls._from_data(read_YAML(string))

    @classmethod
    def from_YAML_file(cls, path: PathLike) -> WorkflowTemplate:
        """Load from a YAML file.

        Parameters
        ----------
        path
            The path to the YAML file containing the workflow template parametrisation.

        """
        return cls._from_data(read_YAML_file(path))

    @classmethod
    def from_JSON_string(cls, string: str) -> WorkflowTemplate:
        """Load from a JSON string.

        Parameters
        ----------
        string
            The JSON string containing the workflow template parametrisation.

        """
        return cls._from_data(read_JSON_string(string))

    @classmethod
    def from_JSON_file(cls, path: PathLike) -> WorkflowTemplate:
        """Load from a JSON file.

        Parameters
        ----------
        path
            The path to the JSON file containing the workflow template parametrisation.

        """
        return cls._from_data(read_JSON_file(path))

    @classmethod
    def from_file(
        cls,
        path: PathLike,
        template_format: Optional[str] = DEFAULT_TEMPLATE_FORMAT,
    ) -> WorkflowTemplate:
        """Load from either a YAML or JSON file, depending on the file extension.

        Parameters
        ----------
        path
            The path to the file containing the workflow template parametrisation.
        template_format
            The file format to expect at `path`. One of "json" or "yaml", if specified. By
            default, "yaml".

        """
        path = Path(path)
        fmt = template_format.lower()
        if fmt == "yaml" or path.suffix in (".yaml", ".yml"):
            return cls.from_YAML_file(path)
        elif fmt == "json" or path.suffix == ".json":
            return cls.from_JSON_file(path)
        else:
            raise ValueError(
                f"Unknown workflow template file extension {path.suffix!r}. Supported "
                f"template formats are {ALL_TEMPLATE_FORMATS!r}."
            )

    def _add_empty_task(self, task: Task, new_index: int, insert_ID: int) -> None:
        """Called by `Workflow._add_empty_task`."""
        new_task_name = self.workflow._get_new_task_unique_name(task, new_index)

        task._insert_ID = insert_ID
        task._dir_name = f"task_{task.insert_ID}_{new_task_name}"
        task._element_sets = []  # element sets are added to the Task during add_elements

        task.workflow_template = self
        self.tasks.insert(new_index, task)

    def _add_empty_loop(self, loop: Loop) -> None:
        """Called by `Workflow._add_empty_loop`."""

        if not loop.name:
            existing = [i.name for i in self.loops]
            new_idx = len(self.loops)
            name = f"loop_{new_idx}"
            while name in existing:
                new_idx += 1
                name = f"loop_{new_idx}"
            loop._name = name
        elif loop.name in self.workflow.loops.list_attrs():
            raise LoopAlreadyExistsError(
                f"A loop with the name {loop.name!r} already exists in the workflow: "
                f"{getattr(self.workflow.loops, loop.name)!r}."
            )

        loop._workflow_template = self
        self.loops.append(loop)


class Workflow:
    """Class to represent a persistent {app_name} workflow."""

    _app_attr = "app"

    _default_ts_fmt = r"%Y-%m-%d %H:%M:%S.%f"
    _default_ts_name_fmt = r"%Y-%m-%d_%H%M%S"

    def __init__(self, path: PathLike) -> None:

        self.path = Path(path).resolve()
        if not self.path.is_dir():
            raise WorkflowNotFoundError(f"No workflow found at path: {self.path}")

        # assigned on first access to corresponding properties:
        self._ts_fmt = None
        self._ts_name_fmt = None
        self._creation_info = None

        self._template = None
        self._template_components = None
        self._tasks = None
        self._loops = None
        self._submissions = None

        self._store = store_cls_from_path(self.path)(self)

        self._in_batch_mode = False  # flag to track when processing batch updates

        # store indices of updates during batch update, so we can revert on failure:
        self._pending = self._get_empty_pending()

    @property
    def name(self):
        """The workflow name may be different from the template name, as it includes the
        creation date-timestamp if generated."""
        return self.path.parts[-1]

    def _get_empty_pending(self) -> Dict:
        return {
            "template_components": {k: [] for k in self.app._template_component_types},
            "tasks": [],  # list of int
            "loops": [],  # list of int
            "submissions": [],  # list of int
        }

    def _accept_pending(self) -> None:
        self._reset_pending()

    def _reset_pending(self) -> None:
        self._pending = self._get_empty_pending()

    def _reject_pending(self) -> None:
        """Revert pending changes to the in-memory representation of the workflow.

        This deletes new tasks, new template component data, new loops, and new
        submissions. Element additions to existing (non-pending) tasks are separately
        rejected/accepted by the WorkflowTask object.

        """
        for task_idx in self._pending["tasks"][::-1]:
            # iterate in reverse so the index references are correct
            self.tasks._remove_object(task_idx)
            self.template.tasks.pop(task_idx)

        for comp_type, comp_indices in self._pending["template_components"].items():
            for comp_idx in comp_indices[::-1]:
                # iterate in reverse so the index references are correct
                self.template_components[comp_type]._remove_object(comp_idx)

        for loop_idx in self._pending["loops"][::-1]:
            # iterate in reverse so the index references are correct
            self.loops._remove_object(loop_idx)
            self.template.loops.pop(loop_idx)

        for sub_idx in self._pending["submissions"][::-1]:
            # iterate in reverse so the index references are correct
            self._submissions.pop(sub_idx)

        self._reset_pending()

    @property
    def store_format(self):
        return self._store.store_name

    @classmethod
    def from_template(
        cls,
        template: WorkflowTemplate,
        path: Optional[PathLike] = None,
        name: Optional[str] = None,
        overwrite: Optional[bool] = False,
        store: Optional[str] = DEFAULT_STORE_FORMAT,
        ts_fmt: Optional[str] = None,
        ts_name_fmt: Optional[str] = None,
    ) -> Workflow:
        """Generate from a `WorkflowTemplate` object.

        Parameters
        ----------
        template
            The WorkflowTemplate object to make persistent.
        path
            The directory in which the workflow will be generated. The current directory
            if not specified.
        name
            The name of the workflow. If specified, the workflow directory will be `path`
            joined with `name`. If not specified the `WorkflowTemplate` name will be used,
            in combination with a date-timestamp.
        overwrite
            If True and the workflow directory (`path` + `name`) already exists, the
            existing directory will be overwritten.
        store
            The persistent store to use for this workflow.
        ts_fmt
            The datetime format to use for storing datetimes. Datetimes are always stored
            in UTC (because Numpy does not store time zone info), so this should not
            include a time zone name.
        ts_name_fmt
            The datetime format to use when generating the workflow name, where it
            includes a timestamp.
        """
        wk = cls._write_empty_workflow(
            template=template,
            path=path,
            name=name,
            overwrite=overwrite,
            store=store,
            ts_fmt=ts_fmt,
            ts_name_fmt=ts_name_fmt,
        )
        with wk._store.cached_load():
            with wk.batch_update(is_workflow_creation=True):
                for task in template.tasks:
                    wk._add_task(task)
                for loop in template.loops:
                    wk._add_loop(loop)
        return wk

    @classmethod
    def from_YAML_file(
        cls,
        YAML_path: PathLike,
        path: Optional[str] = None,
        name: Optional[str] = None,
        overwrite: Optional[bool] = False,
        store: Optional[str] = DEFAULT_STORE_FORMAT,
        ts_fmt: Optional[str] = None,
        ts_name_fmt: Optional[str] = None,
    ) -> Workflow:
        """Generate from a YAML file.

        Parameters
        ----------
        YAML_path
            The path to a workflow template in the YAML file format.
        path
            The directory in which the workflow will be generated. The current directory
            if not specified.
        name
            The name of the workflow. If specified, the workflow directory will be `path`
            joined with `name`. If not specified the `WorkflowTemplate` name will be used,
            in combination with a date-timestamp.
        overwrite
            If True and the workflow directory (`path` + `name`) already exists, the
            existing directory will be overwritten.
        store
            The persistent store to use for this workflow.
        ts_fmt
            The datetime format to use for storing datetimes. Datetimes are always stored
            in UTC (because Numpy does not store time zone info), so this should not
            include a time zone name.
        ts_name_fmt
            The datetime format to use when generating the workflow name, where it
            includes a timestamp.
        """
        template = cls.app.WorkflowTemplate.from_YAML_file(YAML_path)
        return cls.from_template(
            template,
            path,
            name,
            overwrite,
            store,
            ts_fmt,
            ts_name_fmt,
        )

    @classmethod
    def from_YAML_string(
        cls,
        YAML_str: PathLike,
        path: Optional[str] = None,
        name: Optional[str] = None,
        overwrite: Optional[bool] = False,
        store: Optional[str] = DEFAULT_STORE_FORMAT,
        ts_fmt: Optional[str] = None,
        ts_name_fmt: Optional[str] = None,
    ) -> Workflow:
        """Generate from a YAML string.

        Parameters
        ----------
        YAML_str
            The YAML string containing a workflow template parametrisation.
        path
            The directory in which the workflow will be generated. The current directory
            if not specified.
        name
            The name of the workflow. If specified, the workflow directory will be `path`
            joined with `name`. If not specified the `WorkflowTemplate` name will be used,
            in combination with a date-timestamp.
        overwrite
            If True and the workflow directory (`path` + `name`) already exists, the
            existing directory will be overwritten.
        store
            The persistent store to use for this workflow.
        ts_fmt
            The datetime format to use for storing datetimes. Datetimes are always stored
            in UTC (because Numpy does not store time zone info), so this should not
            include a time zone name.
        ts_name_fmt
            The datetime format to use when generating the workflow name, where it
            includes a timestamp.
        """
        template = cls.app.WorkflowTemplate.from_YAML_string(YAML_str)
        return cls.from_template(
            template,
            path,
            name,
            overwrite,
            store,
            ts_fmt,
            ts_name_fmt,
        )

    @classmethod
    def from_JSON_file(
        cls,
        JSON_path: PathLike,
        path: Optional[str] = None,
        name: Optional[str] = None,
        overwrite: Optional[bool] = False,
        store: Optional[str] = DEFAULT_STORE_FORMAT,
        ts_fmt: Optional[str] = None,
        ts_name_fmt: Optional[str] = None,
    ) -> Workflow:
        """Generate from a JSON file.

        Parameters
        ----------
        JSON_path
            The path to a workflow template in the JSON file format.
        path
            The directory in which the workflow will be generated. The current directory
            if not specified.
        name
            The name of the workflow. If specified, the workflow directory will be `path`
            joined with `name`. If not specified the `WorkflowTemplate` name will be used,
            in combination with a date-timestamp.
        overwrite
            If True and the workflow directory (`path` + `name`) already exists, the
            existing directory will be overwritten.
        store
            The persistent store to use for this workflow.
        ts_fmt
            The datetime format to use for storing datetimes. Datetimes are always stored
            in UTC (because Numpy does not store time zone info), so this should not
            include a time zone name.
        ts_name_fmt
            The datetime format to use when generating the workflow name, where it
            includes a timestamp.
        """
        template = cls.app.WorkflowTemplate.from_JSON_file(JSON_path)
        return cls.from_template(
            template,
            path,
            name,
            overwrite,
            store,
            ts_fmt,
            ts_name_fmt,
        )

    @classmethod
    def from_JSON_string(
        cls,
        JSON_str: PathLike,
        path: Optional[str] = None,
        name: Optional[str] = None,
        overwrite: Optional[bool] = False,
        store: Optional[str] = DEFAULT_STORE_FORMAT,
        ts_fmt: Optional[str] = None,
        ts_name_fmt: Optional[str] = None,
    ) -> Workflow:
        """Generate from a JSON string.

        Parameters
        ----------
        JSON_str
            The JSON string containing a workflow template parametrisation.
        path
            The directory in which the workflow will be generated. The current directory
            if not specified.
        name
            The name of the workflow. If specified, the workflow directory will be `path`
            joined with `name`. If not specified the `WorkflowTemplate` name will be used,
            in combination with a date-timestamp.
        overwrite
            If True and the workflow directory (`path` + `name`) already exists, the
            existing directory will be overwritten.
        store
            The persistent store to use for this workflow.
        ts_fmt
            The datetime format to use for storing datetimes. Datetimes are always stored
            in UTC (because Numpy does not store time zone info), so this should not
            include a time zone name.
        ts_name_fmt
            The datetime format to use when generating the workflow name, where it
            includes a timestamp.
        """
        template = cls.app.WorkflowTemplate.from_JSON_string(JSON_str)
        return cls.from_template(
            template,
            path,
            name,
            overwrite,
            store,
            ts_fmt,
            ts_name_fmt,
        )

    @classmethod
    def from_file(
        cls,
        template_path: PathLike,
        template_format: Optional[str] = None,
        path: Optional[str] = None,
        name: Optional[str] = None,
        overwrite: Optional[bool] = False,
        store: Optional[str] = DEFAULT_STORE_FORMAT,
        ts_fmt: Optional[str] = None,
        ts_name_fmt: Optional[str] = None,
    ) -> Workflow:
        """Generate from either a YAML or JSON file, depending on the file extension.

        Parameters
        ----------
        template_path
            The path to a template file in YAML or JSON format, and with a ".yml",
            ".yaml", or ".json" extension.
        template_format
            If specified, one of "json" or "yaml". This forces parsing from a particular
            format regardless of the file extension.
        path
            The directory in which the workflow will be generated. The current directory
            if not specified.
        name
            The name of the workflow. If specified, the workflow directory will be `path`
            joined with `name`. If not specified the `WorkflowTemplate` name will be used,
            in combination with a date-timestamp.
        overwrite
            If True and the workflow directory (`path` + `name`) already exists, the
            existing directory will be overwritten.
        store
            The persistent store to use for this workflow.
        ts_fmt
            The datetime format to use for storing datetimes. Datetimes are always stored
            in UTC (because Numpy does not store time zone info), so this should not
            include a time zone name.
        ts_name_fmt
            The datetime format to use when generating the workflow name, where it
            includes a timestamp.
        """
        template = cls.app.WorkflowTemplate.from_file(template_path, template_format)
        return cls.from_template(
            template,
            path,
            name,
            overwrite,
            store,
            ts_fmt,
            ts_name_fmt,
        )

    @classmethod
    def from_template_data(
        cls,
        template_name: str,
        tasks: Optional[List[Task]] = field(default_factory=lambda: []),
        loops: Optional[List[Loop]] = field(default_factory=lambda: []),
        resources: Optional[Dict[str, Dict]] = None,
        path: Optional[PathLike] = None,
        workflow_name: Optional[str] = None,
        overwrite: Optional[bool] = False,
        store: Optional[str] = DEFAULT_STORE_FORMAT,
        ts_fmt: Optional[str] = None,
        ts_name_fmt: Optional[str] = None,
    ) -> Workflow:
        """Generate from the data associated with a WorkflowTemplate object.

        Parameters
        ----------
        template_name
            Name of the new workflow template, from which the new workflow will be
            generated.
        tasks
            List of Task objects to add to the new workflow.
        loops
            List of Loop objects to add to the new workflow.
        resources
            Mapping of action scopes to resource requirements, to be applied to all
            element sets in the workflow. `resources` specified in an element set take
            precedence of those defined here for the whole workflow.
        path
            The directory in which the workflow will be generated. The current directory
            if not specified.
        workflow_name
            The name of the workflow. If specified, the workflow directory will be `path`
            joined with `name`. If not specified `template_name` will be used, in
            combination with a date-timestamp.
        overwrite
            If True and the workflow directory (`path` + `name`) already exists, the
            existing directory will be overwritten.
        store
            The persistent store to use for this workflow.
        ts_fmt
            The datetime format to use for storing datetimes. Datetimes are always stored
            in UTC (because Numpy does not store time zone info), so this should not
            include a time zone name.
        ts_name_fmt
            The datetime format to use when generating the workflow name, where it
            includes a timestamp.
        """
        template = cls.app.WorkflowTemplate(
            template_name,
            tasks=tasks,
            loops=loops,
            resources=resources,
        )
        return cls.from_template(
            template,
            path,
            workflow_name,
            overwrite,
            store,
            ts_fmt,
            ts_name_fmt,
        )

    @contextmanager
    def batch_update(self, is_workflow_creation: bool = False) -> Iterator[None]:
        """A context manager that batches up structural changes to the workflow and
        commits them to disk all together when the context manager exits."""

        if self._in_batch_mode:
            yield
        else:
            try:
                self._in_batch_mode = True
                yield

            except Exception as err:

                print("batch update exception!")

                self._in_batch_mode = False
                self._store.reject_pending()

                for task in self.tasks:
                    task._reset_pending_elements()

                for loop in self.loops:
                    loop._reset_pending_num_added_iters()

                self._reject_pending()

                if is_workflow_creation:
                    # creation failed, so no need to keep the newly generated workflow:
                    self._store.delete_no_confirm()
                    self._store.reinstate_replaced_dir()

                raise err

            else:

                if self._store.has_pending:

                    is_diff = self._store.is_modified_on_disk()
                    if is_diff:
                        raise WorkflowBatchUpdateFailedError(
                            f"Workflow modified on disk since it was loaded!"
                        )

                    for task in self.tasks:
                        task._accept_pending_elements()

                    for loop in self.loops:
                        loop._accept_pending_num_added_iters()

                    self._store.remove_replaced_dir()
                    # TODO: handle errors in commit pending?
                    self._store.commit_pending()
                    self._accept_pending()
                    self._in_batch_mode = False

    @classmethod
    def _write_empty_workflow(
        cls,
        template: WorkflowTemplate,
        path: Optional[PathLike] = None,
        name: Optional[str] = None,
        overwrite: Optional[bool] = False,
        store: Optional[str] = DEFAULT_STORE_FORMAT,
        ts_fmt: Optional[str] = None,
        ts_name_fmt: Optional[str] = None,
    ) -> Workflow:

        """
        Parameters
        ----------
        path
            The directory in which the workflow will be generated. The current directory
            if not specified.
        name
            The name of the workflow. If specified, the workflow directory will be `path`
            joined with `name`. If not specified the WorkflowTemplate name will be used,
            in combination with a date-timestamp.
        overwrite
            If True and the workflow directory (`path` + `name`) already exists, the
            existing directory will be overwritten.
        store
            The persistent store to use for this workflow.
        ts_fmt
            The datetime format to use for storing datetimes. Datetimes are always stored
            in UTC (because Numpy does not store time zone info), so this should not
            include a time zone name.
        ts_name_fmt
            The datetime format to use when generating the workflow name, where it
            includes a timestamp.
        """

        ts = datetime.now()

        # store all times in UTC, since Numpy doesn't support time zone info:
        ts_utc = ts.astimezone(tz=timezone.utc)

        ts_name_fmt = ts_name_fmt or cls._default_ts_name_fmt
        ts_fmt = ts_fmt or cls._default_ts_fmt

        path = Path(path or "").resolve()
        name = name or f"{template.name}_{ts.strftime(ts_name_fmt)}"
        workflow_path = path.joinpath(name)

        replaced_dir = None
        if workflow_path.exists():
            if overwrite:
                replaced_dir = temporary_workflow_rename(workflow_path)
            else:
                raise ValueError(f"Path already exists: {workflow_path}.")

        # make template-level inputs/resources think they are persistent:
        wk_dummy = _DummyPersistentWorkflow()
        param_src = {"type": "workflow_resources"}
        for res_i in template.resources:
            res_i.make_persistent(wk_dummy, param_src)

        template_js, template_sh = template.to_json_like(exclude=["tasks", "loops"])
        template_js["tasks"] = []
        template_js["loops"] = []

        creation_info = {
            "app_info": cls.app.get_info(),
            "create_time": ts_utc.strftime(ts_fmt),
            "ts_fmt": ts_fmt,
            "ts_name_fmt": ts_name_fmt,
        }

        store_cls = store_cls_from_str(store)
        store_cls.write_empty_workflow(
            template_js=template_js,
            template_components_js=template_sh,
            workflow_path=workflow_path,
            replaced_dir=replaced_dir,
            creation_info=creation_info,
        )
        wk = cls(workflow_path)

        # actually make template inputs/resources persistent, now the workflow exists:
        wk_dummy.make_persistent(wk)

        return wk

    @property
    def ts_fmt(self):
        if not self._ts_fmt:
            self._ts_fmt = self._store.get_creation_info()["ts_fmt"]
        return self._ts_fmt

    @property
    def ts_name_fmt(self):
        if not self._ts_name_fmt:
            self._ts_name_fmt = self._store.get_creation_info()["ts_name_fmt"]
        return self._ts_name_fmt

    @property
    def creation_info(self):
        if not self._creation_info:
            with self._store.cached_load():
                info = self._store.get_creation_info()
                info["create_time"] = datetime.strptime(
                    info["create_time"], info["ts_fmt"]
                ).replace(tzinfo=timezone.utc)
                self._creation_info = info
        return self._creation_info

    @property
    def num_tasks(self) -> int:
        return len(self.tasks)

    @property
    def num_added_tasks(self) -> int:
        with self._store.cached_load():
            return self._store.get_num_added_tasks()

    @property
    def num_elements(self) -> int:
        return sum(task.num_elements for task in self.tasks)

    @property
    def num_element_iterations(self) -> int:
        return sum(task.num_element_iterations for task in self.tasks)

    @property
    def num_loops(self) -> int:
        return len(self.loops)

    @property
    def num_submissions(self) -> int:
        return len(self.submissions)

    @property
    def template_components(self) -> Dict:
        if self._template_components is None:
            with self._store.cached_load():
                tc_js = self._store.get_template_components()
            self._template_components = self.app.template_components_from_json_like(tc_js)
        return self._template_components

    @property
    def template(self) -> WorkflowTemplate:
        if self._template is None:
            with self._store.cached_load():
                temp_js = self._store.get_template()
                template = self.app.WorkflowTemplate.from_json_like(
                    temp_js, self.template_components
                )
                template.workflow = self
            self._template = template

        return self._template

    @property
    def tasks(self) -> WorkflowTaskList:
        if self._tasks is None:
            with self._store.cached_load():
                tasks_meta = self._store.get_all_tasks_metadata()
                wk_tasks = []
                for idx, i in enumerate(tasks_meta):
                    wk_task = self.app.WorkflowTask(
                        workflow=self,
                        template=self.template.tasks[idx],
                        index=idx,
                        num_elements=i["num_elements"],
                        num_element_iterations=i["num_element_iterations"],
                        num_EARs=i["num_EARs"],
                    )
                    wk_tasks.append(wk_task)
                self._tasks = self.app.WorkflowTaskList(wk_tasks)
        return self._tasks

    @property
    def loops(self) -> WorkflowLoopList:
        if self._loops is None:
            with self._store.cached_load():
                wk_loops = []
                for idx, loop_dat in enumerate(self._store.get_loops()):
                    wk_loop = self.app.WorkflowLoop(
                        index=idx,
                        workflow=self,
                        template=self.template.loops[idx],
                        **loop_dat,
                    )
                    wk_loops.append(wk_loop)
                self._loops = self.app.WorkflowLoopList(wk_loops)
        return self._loops

    @property
    def submissions(self) -> List[Submission]:
        if self._submissions is None:
            with self._store.cached_load():
                subs = []
                for idx, sub_dat in enumerate(self._store.get_submissions()):
                    sub_js = {"index": idx, "workflow": self, **sub_dat}
                    sub = self.app.Submission.from_json_like(sub_js)
                    subs.append(sub)
                self._submissions = subs
        return self._submissions

    @property
    def artifacts_path(self):
        # TODO: allow customisation of artifacts path at submission and resources level
        return self.path / "artifacts"

    @property
    def submissions_path(self):
        return self.artifacts_path / "submissions"

    @property
    def task_artifacts_path(self):
        return self.artifacts_path / "tasks"

    def elements(self) -> Iterator[Element]:
        for task in self.tasks:
            for element in task.elements:
                yield element

    def copy(self, path=None) -> Workflow:
        """Copy the workflow to a new path and return the copied workflow."""
        if path is None:
            path = self.path.parent / Path(self.path.stem + "_copy" + self.path.suffix)
        if path.exists():
            raise ValueError(f"Path already exists: {path}.")
        self._store.copy(path=path)
        return self.app.Workflow(path=path)

    def delete(self):
        self._store.delete()

    def _delete_no_confirm(self):
        self._store.delete_no_confirm()

    def rename(self, new_name: str):
        raise NotImplementedError

    def _submit(
        self,
        ignore_errors: Optional[bool] = False,
        JS_parallelism: Optional[bool] = None,
    ):
        """Submit outstanding EARs for execution."""

        # generate a new submission if there are no pending submissions:
        pending = [i for i in self.submissions if i.needs_submit]
        if not pending:
            new_sub = self.add_submission(JS_parallelism=JS_parallelism)
            if not new_sub:
                raise ValueError("No pending element action runs to submit!")
            pending = [new_sub]

        self.submissions_path.mkdir(exist_ok=True, parents=True)
        self.task_artifacts_path.mkdir(exist_ok=True, parents=True)

        # for direct execution the submission must be persistent at submit-time, because
        # it will be read by a new instance of the app:
        self._store.commit_pending()

        # submit all pending submissions:
        exceptions = []
        submitted_js = {}
        for sub in pending:
            try:
                sub_js_idx = sub.submit(
                    self.task_artifacts_path,
                    ignore_errors=ignore_errors,
                )
                submitted_js[sub.index] = sub_js_idx
            except SubmissionFailure as exc:
                exceptions.append(exc)

        return exceptions, submitted_js

    def submit(
        self,
        ignore_errors: Optional[bool] = False,
        JS_parallelism: Optional[bool] = None,
    ) -> None:
        with self._store.cached_load():
            with self.batch_update():
                # commit updates before raising exception:
                exceptions, submitted_js = self._submit(
                    ignore_errors=ignore_errors,
                    JS_parallelism=JS_parallelism,
                )

        if exceptions:
            msg = "\n" + "\n\n".join([i.message for i in exceptions])
            raise WorkflowSubmissionFailure(msg)

        return submitted_js

    def add_submission(self, JS_parallelism: Optional[bool] = None) -> Submission:
        new_idx = self.num_submissions
        sub_obj = self.app.Submission(
            index=new_idx,
            workflow=self,
            jobscripts=self.resolve_jobscripts(),
            JS_parallelism=JS_parallelism,
        )
        EAR_indices = sub_obj.prepare_EAR_submission_idx_update()
        if not EAR_indices:
            print(
                f"There are no pending element action runs, so a new submission was not "
                f"added."
            )
            return

        self.set_EAR_submission_indices(sub_idx=new_idx, EAR_indices=EAR_indices)
        sub_obj_js, _ = sub_obj.to_json_like()
        self._submissions.append(sub_obj)
        self._pending["submissions"].append(new_idx)
        with self._store.cached_load():
            with self.batch_update():
                self._store.add_submission(sub_obj_js)

        return self.submissions[new_idx]

    def get_task_unique_names(
        self, map_to_insert_ID: bool = False
    ) -> Union[List[str], Dict[str, int]]:
        """Return the unique names of all workflow tasks.

        Parameters
        ----------
        map_to_insert_ID : bool, optional
            If True, return a dict whose values are task insert IDs, otherwise return a
            list.

        """
        names = Task.get_task_unique_names(self.template.tasks)
        if map_to_insert_ID:
            insert_IDs = (i.insert_ID for i in self.template.tasks)
            return dict(zip(names, insert_IDs))
        else:
            return names

    def _get_new_task_unique_name(self, new_task: Task, new_index: int) -> str:

        task_templates = list(self.template.tasks)
        task_templates.insert(new_index, new_task)
        uniq_names = Task.get_task_unique_names(task_templates)

        return uniq_names[new_index]

    def _add_empty_task(
        self,
        task: Task,
        new_index: Optional[int] = None,
    ) -> WorkflowTask:

        if new_index is None:
            new_index = self.num_tasks

        insert_ID = self.num_added_tasks

        # make a copy with persistent schema inputs:
        task_c, _ = task.to_persistent(self, insert_ID)

        # add to the WorkflowTemplate:
        self.template._add_empty_task(task_c, new_index, insert_ID)

        # create and insert a new WorkflowTask:
        self.tasks.add_object(
            self.app.WorkflowTask.new_empty_task(self, task_c, new_index),
            index=new_index,
        )

        # update persistent store:
        task_js, temp_comps_js = task_c.to_json_like()
        self._store.add_template_components(temp_comps_js)
        self._store.add_empty_task(new_index, task_js)

        # update in-memory workflow template components:
        temp_comps = self.app.template_components_from_json_like(temp_comps_js)
        for comp_type, comps in temp_comps.items():
            for comp in comps:
                comp._set_hash()
                if comp not in self.template_components[comp_type]:
                    idx = self.template_components[comp_type].add_object(comp)
                    self._pending["template_components"][comp_type].append(idx)

        self._pending["tasks"].append(new_index)

        return self.tasks[new_index]

    def _add_empty_loop(self, loop: Loop) -> WorkflowLoop:
        """Add a new loop (zeroth iterations only) to the workflow."""

        new_index = self.num_loops

        # don't modify passed object:
        loop_c = copy.deepcopy(loop)

        # add to the WorkflowTemplate:
        self.template._add_empty_loop(loop_c)

        # create and insert a new WorkflowLoop:
        self.loops.add_object(
            self.app.WorkflowLoop.new_empty_loop(
                index=new_index,
                workflow=self,
                template=loop_c,
            )
        )
        wk_loop = self.loops[new_index]

        # update persistent store:
        loop_js, _ = loop_c.to_json_like()
        task_indices = [self.tasks.get(insert_ID=i).index for i in loop_c.task_insert_IDs]
        self._store.add_loop(
            task_indices=task_indices,
            loop_js=loop_js,
            iterable_parameters=wk_loop.iterable_parameters,
        )

        self._pending["loops"].append(new_index)

        return wk_loop

    def _add_loop(self, loop: Loop, parent_loop_indices: Dict = None) -> None:
        new_wk_loop = self._add_empty_loop(loop)
        if loop.num_iterations is not None:
            # fixed number of iterations, so add remaining N > 0 iterations:
            for _ in range(loop.num_iterations - 1):
                new_wk_loop.add_iteration(parent_loop_indices=parent_loop_indices)

    def add_loop(self, loop: Loop, parent_loop_indices: Dict = None) -> None:
        """Add a loop to a subset of workflow tasks."""
        with self._store.cached_load():
            with self.batch_update():
                self._add_loop(loop, parent_loop_indices)

    def _add_task(self, task: Task, new_index: Optional[int] = None) -> None:
        new_wk_task = self._add_empty_task(task=task, new_index=new_index)
        new_wk_task._add_elements(element_sets=task.element_sets)

    def add_task(self, task: Task, new_index: Optional[int] = None) -> None:
        with self._store.cached_load():
            with self.batch_update():
                self._add_task(task, new_index=new_index)

    def add_task_after(self, new_task: Task, task_ref=None):
        """Adds the given new_task after the task specified in task_ref.

        Parameters
        ----------
        new_task : Task, required.
        task_ref : Task, optional.
            If no `task_ref` is given, the new task will be added at the end.

        """
        new_index = task_ref.index + 1 if task_ref else None
        self.add_task(new_task, new_index)
        # TODO: add new downstream elements?

    def add_task_before(self, new_task: Task, task_ref=None):
        """Adds the given new_task before the task specified in task_ref.

        Parameters
        ----------
        new_task : Task, required.
        task_ref : Task, optional.
            If no task_ref is given, the new task will be added at the beginning.

        """
        new_index = task_ref.index if task_ref else 0
        self.add_task(new_task, new_index)
        # TODO: add new downstream elements?

    def get_parameter_data(self, index: int) -> Tuple[bool, Any]:
        return self._store.get_parameter_data(index)

    def get_parameter_source(self, index: int) -> Dict:
        return self._store.get_parameter_source(index)

    def get_all_parameter_data(self) -> Dict[int, Any]:
        return self._store.get_all_parameter_data()

    def is_parameter_set(self, index: int) -> bool:
        return self._store.is_parameter_set(index)

    def check_parameters_exist(
        self, indices: Union[int, List[int]]
    ) -> Union[bool, List[bool]]:
        return self._store.check_parameters_exist(indices)

    def _add_unset_parameter_data(self, source: Dict) -> int:
        return self._store.add_unset_parameter_data(source)

    def _add_parameter_data(self, data, source: Dict) -> int:
        return self._store.add_parameter_data(data, source)

    def _resolve_input_source_task_reference(
        self, input_source: InputSource, new_task_name: str
    ) -> None:
        """Normalise the input source task reference and convert a source to a local type
        if required."""

        # TODO: test thoroughly!

        if isinstance(input_source.task_ref, str):
            if input_source.task_ref == new_task_name:
                if input_source.task_source_type is self.app.TaskSourceType.OUTPUT:
                    raise InvalidInputSourceTaskReference(
                        f"Input source {input_source.to_string()!r} cannot refer to the "
                        f"outputs of its own task!"
                    )
                else:
                    warn(
                        f"Changing input source {input_source.to_string()!r} to a local "
                        f"type, since the input source task reference refers to its own "
                        f"task."
                    )
                    # TODO: add an InputSource source_type setter to reset
                    # task_ref/source_type?
                    input_source.source_type = self.app.InputSourceType.LOCAL
                    input_source.task_ref = None
                    input_source.task_source_type = None
            else:
                try:
                    uniq_names_cur = self.get_task_unique_names(map_to_insert_ID=True)
                    input_source.task_ref = uniq_names_cur[input_source.task_ref]
                except KeyError:
                    raise InvalidInputSourceTaskReference(
                        f"Input source {input_source.to_string()!r} refers to a missing "
                        f"or inaccessible task: {input_source.task_ref!r}."
                    )

    def get_task_elements(self, task: Task, selection: slice) -> List[Element]:
        return [
            self.app.Element(task=task, **i)
            for i in self._store.get_task_elements(task.index, task.insert_ID, selection)
        ]

    def get_task_elements_islice(self, task: Task, selection: slice) -> Iterator[Element]:
        for i in self._store.get_task_elements_islice(
            task.index, task.insert_ID, selection
        ):
            yield self.app.Element(task=task, **i)

    def get_EARs_from_IDs(self, indices: List[EAR_ID]) -> List[ElementActionRun]:
        """Return element action run objects from a list of five-tuples, representing the
        task insert ID, element index, iteration index, action index, and run index,
        respectively.
        """
        objs = []
        for _EAR_ID in indices:
            task = self.tasks.get(insert_ID=_EAR_ID.task_insert_ID)
            elem_iters = task.elements[_EAR_ID.element_idx].iterations
            for i in elem_iters:
                if i.index == _EAR_ID.iteration_idx:
                    iter_i = i
                    break
            EAR_i = iter_i.actions[_EAR_ID.action_idx].runs[_EAR_ID.run_idx]
            objs.append(EAR_i)
        return objs

    def get_element_iterations_from_IDs(
        self, indices: List[IterationID]
    ) -> List[ElementIteration]:
        """Return element iteration objects from a list of three-tuples, representing the
        task insert ID, element index, and iteration index, respectively.
        """
        objs = []
        for iter_idx in indices:
            iter_i = (
                self.tasks.get(insert_ID=iter_idx.task_insert_ID)
                .elements[iter_idx.element_idx]
                .iterations[iter_idx.iteration_idx]
            )
            objs.append(iter_i)
        return objs

    def get_elements_from_IDs(self, indices: List[ElementID]) -> List[Element]:
        """Return element objects from a list of two-tuples, representing the task insert
        ID, and element index, respectively."""
        return [
            self.tasks.get(insert_ID=idx.task_insert_ID).elements[idx.element_idx]
            for idx in indices
        ]

    def set_EAR_submission_indices(
        self,
        sub_idx: int,
        EAR_indices: Tuple[int, int, int, int],
    ) -> None:
        """Set the submission index on an EAR."""
        with self._store.cached_load():
            with self.batch_update():
                self._store.set_EAR_submission_indices(sub_idx, EAR_indices)

    def set_EAR_start(
        self,
        submission_idx: int,
        jobscript_idx: int,
        JS_element_idx: int,
        JS_action_idx: int,
    ) -> None:
        """Set the start time on an EAR."""
        with self._store.cached_load():
            with self.batch_update():
                jobscript = self.submissions[submission_idx].jobscripts[jobscript_idx]
                (t_iD, _, i_idx, a_idx, r_idx, _) = jobscript.get_EAR_ID_array()[
                    JS_action_idx, JS_element_idx
                ].item()
                self._store.set_EAR_start(t_iD, i_idx, a_idx, r_idx)

    def set_EAR_end(
        self,
        submission_idx: int,
        jobscript_idx: int,
        JS_element_idx: int,
        JS_action_idx: int,
    ) -> None:
        """Set the end time on an EAR."""
        with self._store.cached_load():
            with self.batch_update():
                jobscript = self.submissions[submission_idx].jobscripts[jobscript_idx]
                (t_iD, _, i_idx, a_idx, r_idx, _) = jobscript.get_EAR_ID_array()[
                    JS_action_idx, JS_element_idx
                ].item()
                self._store.set_EAR_end(t_iD, i_idx, a_idx, r_idx)

    def _from_internal_get_EAR(
        self,
        submission_idx: int,
        jobscript_idx: int,
        JS_element_idx: int,
        JS_action_idx: int,
    ):
        with self._store.cached_load():
            jobscript = self.submissions[submission_idx].jobscripts[jobscript_idx]
            id_args = jobscript.get_EAR_ID_array()[JS_action_idx, JS_element_idx].item()
            EAR_id = EAR_ID(*id_args)
            EAR = self.get_EARs_from_IDs([EAR_id])[0]

        return jobscript, EAR

    def write_commands(
        self,
        submission_idx: int,
        jobscript_idx: int,
        JS_element_idx: int,
        JS_action_idx: int,
    ) -> None:
        """Write run-time commands for a given EAR."""
        with self._store.cached_load():
            jobscript, EAR = self._from_internal_get_EAR(
                submission_idx, jobscript_idx, JS_element_idx, JS_action_idx
            )
            commands, shell_vars = EAR.compose_commands(jobscript)
            for param_name, shell_var_name in shell_vars:
                commands += jobscript.shell.format_save_parameter(
                    workflow_app_alias=jobscript.workflow_app_alias,
                    param_name=param_name,
                    shell_var_name=shell_var_name,
                )
            with Path(jobscript.commands_file_name).open("wt", newline="\n") as fp:
                # (assuming we have CD'd correctly to the element run directory)
                fp.write(commands)

    def save_parameter(
        self,
        name,
        value,
        submission_idx: int,
        jobscript_idx: int,
        JS_element_idx: int,
        JS_action_idx: int,
    ):
        with self._store.cached_load():
            with self.batch_update():
                _, EAR = self._from_internal_get_EAR(
                    submission_idx, jobscript_idx, JS_element_idx, JS_action_idx
                )
                data_idx = EAR.data_idx[name]
                self._store.set_parameter(data_idx, value)

    def resolve_jobscripts(self) -> List[Jobscript]:

        js, element_deps = self._resolve_singular_jobscripts()
        js_deps = resolve_jobscript_dependencies(js, element_deps)

        for js_idx in js:
            if js_idx in js_deps:
                js[js_idx]["dependencies"] = js_deps[js_idx]

        js = merge_jobscripts_across_tasks(js)
        js = jobscripts_to_list(js)
        js_objs = [self.app.Jobscript(**i) for i in js]

        return js_objs

    def _resolve_singular_jobscripts(self) -> Tuple[Dict[int, Dict], Dict]:
        """
        We arrange EARs into `EARs` and `elements` so we can quickly look up membership
        by EAR idx in the `EARs` dict.

        Returns
        -------
        submission_jobscripts
        all_element_deps

        """

        submission_jobscripts = {}
        all_element_deps = {}

        for task_iID, loop_idx_i in self.get_iteration_task_pathway():

            task = self.tasks.get(insert_ID=task_iID)
            res, res_hash, res_map, EAR_map = generate_EAR_resource_map(task, loop_idx_i)
            jobscripts, _ = group_resource_map_into_jobscripts(res_map)

            for js_dat in jobscripts:

                # (insert ID, action_idx, index into task_loop_idx):
                task_actions = [
                    [task.insert_ID, i, 0]
                    for i in sorted(
                        set(
                            act_idx_i
                            for act_idx in js_dat["elements"].values()
                            for act_idx_i in act_idx
                        )
                    )
                ]
                task_elements = {task.insert_ID: list(js_dat["elements"].keys())}
                EAR_idx_arr_shape = (
                    len(task_actions),
                    len(task_elements[task.insert_ID]),
                )
                EAR_idx_arr = np.empty(EAR_idx_arr_shape, dtype=np.int32)
                EAR_idx_arr[:] = -1

                new_js_idx = len(submission_jobscripts)

                js_i = {
                    "task_insert_IDs": [task.insert_ID],
                    "task_loop_idx": [loop_idx_i],
                    "task_actions": task_actions,  # map jobscript actions to task actions
                    "task_elements": task_elements,  # map jobscript elements to task elements
                    "EARs": {},  # keys are (task insert ID, elem_idx, EAR_idx)
                    "EAR_idx": EAR_idx_arr,
                    "resources": res[js_dat["resources"]],
                    "resource_hash": res_hash[js_dat["resources"]],
                    "dependencies": {},
                }
                for elem_idx, act_indices in js_dat["elements"].items():

                    js_elem_idx = task_elements[task.insert_ID].index((elem_idx))
                    all_EAR_IDs = []
                    for act_idx in act_indices:
                        EAR_idx, run_idx, iter_idx = (
                            i.item() for i in EAR_map[act_idx, elem_idx]
                        )
                        # construct EAR_ID object so we can retrieve the EAR objects and
                        # so their dependencies:
                        EAR_id = EAR_ID(
                            task_insert_ID=task.insert_ID,
                            element_idx=elem_idx,
                            iteration_idx=iter_idx,
                            action_idx=act_idx,
                            run_idx=run_idx,
                            EAR_idx=EAR_idx,
                        )
                        all_EAR_IDs.append(EAR_id)
                        js_i["EARs"][(task.insert_ID, elem_idx, EAR_idx)] = (
                            iter_idx,
                            act_idx,
                            run_idx,
                        )

                        js_act_idx = task_actions.index([task.insert_ID, act_idx, 0])
                        js_i["EAR_idx"][js_act_idx][js_elem_idx] = EAR_idx

                    # get indices of EARs that this element depends on:
                    EAR_objs = self.get_EARs_from_IDs(all_EAR_IDs)
                    EAR_deps = [i.get_EAR_dependencies() for i in EAR_objs]
                    EAR_deps_flat = [j for i in EAR_deps for j in i]

                    # represent EAR dependencies of this jobscripts using the same key
                    # format as in the "EARs" dict, to allow for quick lookup when
                    # resolving dependencies between jobscripts; also, no need to include
                    # EAR dependencies that are in this jobscript:
                    EAR_deps_EAR_idx = [
                        (i.task_insert_ID, i.element_idx, i.EAR_idx)
                        for i in EAR_deps_flat
                        if (i.task_insert_ID, i.element_idx, i.EAR_idx)
                        not in js_i["EARs"]
                    ]
                    if EAR_deps_EAR_idx:
                        if new_js_idx not in all_element_deps:
                            all_element_deps[new_js_idx] = {}

                        all_element_deps[new_js_idx][js_elem_idx] = EAR_deps_EAR_idx

                submission_jobscripts[new_js_idx] = js_i

        return submission_jobscripts, all_element_deps

    def get_iteration_task_pathway(self):
        pathway = []
        for task in self.tasks:
            loop_idx = {}
            pathway.append((task.insert_ID, loop_idx))

        for loop in self.loops:  # TODO: order by depth (inner loops first?)
            task_subset = loop.task_insert_IDs
            subset_idx = [idx for idx, i in enumerate(pathway) if i[0] in task_subset]
            looped_pathway = []
            for iter_i in range(loop.num_added_iterations):
                for j in subset_idx:
                    item_j = copy.deepcopy(pathway[j])
                    item_j[1][loop.name] = iter_i
                    looped_pathway.append(item_j)

            # replaced pathway `sub_idx` items with `looped_pathway` items:
            pathway = replace_items(
                pathway, subset_idx[0], subset_idx[-1] + 1, looped_pathway
            )

        return pathway

    def show_all_EAR_statuses(self):
        print(
            f"{'task':8s} {'element':8s} {'iteration':8s} {'action':8s} "
            f"{'run':8s} {'status':8s}"
        )
        for task in self.tasks:
            for element in task.elements:
                for iter_idx, iteration in enumerate(element.iterations):
                    for act_idx, action_runs in iteration.actions.items():
                        for run_idx, EAR in enumerate(action_runs.runs):
                            print(
                                f"{task.insert_ID:^8d} {element.index:^8d} "
                                f"{iter_idx:^8d} {act_idx:^8d} {run_idx:^8d} "
                                f"{EAR.submission_status.name.lower():^8s}"
                            )


@dataclass
class WorkflowBlueprint:
    """Pre-built workflow templates that are simpler to parametrise (e.g. fitting workflows)."""

    workflow_template: WorkflowTemplate



import os
from typing import Dict, Optional

from hpcflow.sdk.core.errors import UnsupportedShellError

from .base import Shell
from .bash import Bash, WSLBash
from .powershell import WindowsPowerShell

ALL_SHELLS = {
    "bash": {"posix": Bash},
    "powershell": {"nt": WindowsPowerShell},
    "wsl+bash": {"nt": WSLBash},
    "wsl": {"nt": WSLBash},
}

DEFAULT_SHELL_NAMES = {
    "posix": "bash",
    "nt": "powershell",
}


def get_supported_shells(os_name: Optional[str] = None) -> Dict[str, Shell]:
    os_name = os_name or os.name
    return {k: v.get(os_name) for k, v in ALL_SHELLS.items() if v.get(os_name)}


def get_shell(shell_name, os_name: Optional[str] = None, **kwargs) -> Shell:
    # TODO: apply config default shell args?

    shell_name = shell_name.lower()

    shell_cls = get_supported_shells(os_name).get(shell_name)
    if not shell_cls:
        raise UnsupportedShellError(f"Shell name {shell_name!r} is not supported.")

    shell_obj = shell_cls(**kwargs)

    return shell_obj



import logging


class AppLog:

    DEFAULT_LOG_LEVEL_CONSOLE = "WARNING"
    DEFAULT_LOG_LEVEL_FILE = "INFO"
    DEFAULT_LOG_FILE_PATH = "app.log"

    def __init__(self, app, log_console_level=None):
        self.app = app
        self.logger = logging.getLogger(app.name)
        self.logger.setLevel(logging.DEBUG)
        self.console_handler = self._add_console_logger(
            level=log_console_level or AppLog.DEFAULT_LOG_LEVEL_CONSOLE
        )
        # print(f"log_console_level: {log_console_level}")

    def _add_console_logger(self, level, fmt=None):
        if not fmt:
            fmt = "%(levelname)s %(name)s: %(message)s"
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(fmt))
        handler.setLevel(level)
        self.logger.addHandler(handler)
        return handler



"""An hpcflow application."""

from functools import wraps
from importlib import resources, import_module
import time
from typing import Dict
import warnings

import click
from colorama import init as colorama_init
from termcolor import colored

from hpcflow import __version__
from hpcflow.sdk.persistence import ALL_STORE_FORMATS, DEFAULT_STORE_FORMAT
from hpcflow.sdk.submission.shells import ALL_SHELLS
from .core.json_like import JSONLike
from .core.utils import read_YAML, read_YAML_file
from . import api, SDK_logger
from .config import Config
from .config.cli import get_config_CLI
from .config.errors import ConfigError
from .core.actions import (
    Action,
    ActionScopeType,
    ElementAction,
    ElementAction,
    ElementActionRun,
)
from .core.element import (
    Element,
    ElementIteration,
    ElementInputFiles,
    ElementInputs,
    ElementOutputFiles,
    ElementOutputs,
    ElementResources,
    ElementParameter,
)
from .core.parameters import (
    InputSourceType,
    ParameterPropagationMode,
    TaskSourceType,
    ValueSequence,
    ParameterValue,
)
from .core.task import (
    ElementPropagation,
    Parameters,
    TaskInputParameters,
    TaskOutputParameters,
    WorkflowTask,
    ElementSet,
    Elements,
)
from .core.loop import Loop, WorkflowLoop
from .core.task_schema import TaskObjective
from .core.workflow import ALL_TEMPLATE_FORMATS, DEFAULT_TEMPLATE_FORMAT, Workflow
from .demo.cli import get_demo_software_CLI
from .helper.cli import get_helper_CLI
from .log import AppLog
from .runtime import RunTimeInfo

SDK_logger = SDK_logger.getChild(__name__)


class BaseApp:
    """Class to generate the base hpcflow application."""

    _template_component_types = (
        "parameters",
        "command_files",
        "environments",
        "task_schemas",
    )

    def __init__(
        self,
        name,
        version,
        description,
        config_options,
        template_components: Dict = None,
        pytest_args=None,
    ):
        SDK_logger.info(f"Generating {self.__class__.__name__} {name!r}.")

        self.name = name
        self.version = version
        self.description = description
        self.config_options = config_options
        self.pytest_args = pytest_args

        self.CLI = self._make_CLI()
        self.log = AppLog(self)
        self.run_time_info = RunTimeInfo(
            self.name, self.version, self.runtime_info_logger
        )

        self._builtin_template_components = template_components or {}

        self._config = None  # assigned on first access to `config` property

        # Set by `_load_template_components`:
        self._template_components = {}
        self._parameters = None
        self._command_files = None
        self._environments = None
        self._task_schemas = None
        self._scripts = None

        self._core_classes = self._assign_core_classes()

        # Add API functions as methods:
        SDK_logger.debug(f"Assigning API functions to the {self.__class__.__name__}.")

        def get_api_method(func):
            # this function avoids scope issues
            return lambda *args, **kwargs: func(self, *args, **kwargs)

        for func in (getattr(api, i) for i in api.__all__):

            if type(self) is BaseApp and func.__name__ == "run_hpcflow_tests":
                # this method provides the same functionality as the `run_tests` method
                continue

            SDK_logger.debug(f"Wrapping API callable: {func!r}")
            # allow sub-classes to override API functions:
            if not hasattr(self, func.__name__):
                api_method = get_api_method(func)
                api_method = wraps(func)(api_method)
                api_method.__doc__ = func.__doc__.format(app_name=name)
                setattr(self, func.__name__, api_method)

    @property
    def template_components(self):
        if not self.is_template_components_loaded:
            self._load_template_components()
        return self._template_components

    def _get_core_JSONLike_classes(self):
        """Get all JSONLike subclasses (recursively).

        If this is run after App initialisation, the returned list will include the
        app-injected sub-classes as well.

        """

        def all_subclasses(cls):
            return set(cls.__subclasses__()).union(
                [s for c in cls.__subclasses__() for s in all_subclasses(c)]
            )

        return all_subclasses(JSONLike)

    def inject_into(self, cls):
        SDK_logger.debug(f"Injecting app {self!r} into class {cls.__name__}")
        if cls.__doc__:
            cls.__doc__ = cls.__doc__.format(app_name=self.name)
        return type(cls.__name__, (cls,), {getattr(cls, "_app_attr"): self})

    def _assign_core_classes(self):

        # ensure classes defined in `object_list` are included in core classes:
        import_module("hpcflow.sdk.core.object_list")

        core_classes = list(self._get_core_JSONLike_classes())

        # Non-`JSONLike` classes:
        core_classes += [
            ActionScopeType,
            Element,
            ElementIteration,
            Elements,
            ElementInputs,
            ElementInputs,
            ElementOutputs,
            ElementResources,
            ElementInputFiles,
            ElementOutputFiles,
            ElementAction,
            ElementAction,
            ElementActionRun,
            ElementParameter,
            ElementPropagation,
            InputSourceType,
            Parameters,
            ParameterPropagationMode,
            TaskSourceType,
            TaskInputParameters,
            TaskOutputParameters,
            Workflow,
            WorkflowTask,
            WorkflowLoop,
            ParameterValue,
        ]
        for cls in core_classes:
            if hasattr(cls, "_app_attr"):
                setattr(self, cls.__name__, self.inject_into(cls))
            else:
                setattr(self, cls.__name__, cls)

        return tuple(
            sorted(
                set(core_classes),
                key=lambda x: f"{x.__module__}.{x.__qualname__}",
            )
        )

    def _ensure_template_components(self):
        if not self.is_template_components_loaded:
            self._load_template_components()

    def load_template_components(self, warn=True):
        if warn and self.is_template_components_loaded:
            warnings.warn("Template components already loaded; reloading now.")
        self._load_template_components()

    def reload_template_components(self, warn=True):
        if warn and not self.is_template_components_loaded:
            warnings.warn("Template components not loaded; loading now.")
        self._load_template_components()

    def _load_template_components(self):
        """Combine any builtin template components with user-defined template components
        and initialise list objects."""

        params = self._builtin_template_components.get("parameters", [])
        for path in self.config.parameter_sources:
            params.extend(read_YAML_file(path))

        cmd_files = self._builtin_template_components.get("command_files", [])
        for path in self.config.command_file_sources:
            cmd_files.extend(read_YAML_file(path))

        envs = self._builtin_template_components.get("environments", [])
        for path in self.config.environment_sources:
            envs.extend(read_YAML_file(path))

        schemas = self._builtin_template_components.get("task_schemas", [])
        for path in self.config.task_schema_sources:
            schemas.extend(read_YAML_file(path))

        self_tc = self._template_components
        self_tc["parameters"] = self.ParametersList.from_json_like(
            params, shared_data=self_tc
        )
        self_tc["command_files"] = self.CommandFilesList.from_json_like(
            cmd_files, shared_data=self_tc
        )
        self_tc["environments"] = self.EnvironmentsList.from_json_like(
            envs, shared_data=self_tc
        )
        self_tc["task_schemas"] = self.TaskSchemasList.from_json_like(
            schemas, shared_data=self_tc
        )

        self._parameters = self_tc["parameters"]
        self._command_files = self_tc["command_files"]
        self._environments = self_tc["environments"]
        self._task_schemas = self_tc["task_schemas"]

        self._scripts = self._load_scripts()

        self.logger.info("Template components loaded.")

    @classmethod
    def load_builtin_template_component_data(cls, package):
        components = {}
        for comp_type in cls._template_component_types:
            with resources.open_text(package, f"{comp_type}.yaml") as fh:
                comp_dat = fh.read()
                components[comp_type] = read_YAML(comp_dat)
        return components

    @property
    def parameters(self):
        self._ensure_template_components()
        return self._parameters

    @property
    def command_files(self):
        self._ensure_template_components()
        return self._command_files

    @property
    def envs(self):
        self._ensure_template_components()
        return self._environments

    @property
    def scripts(self):
        self._ensure_template_components()
        return self._scripts

    @property
    def task_schemas(self):
        self._ensure_template_components()
        return self._task_schemas

    @property
    def logger(self):
        return self.log.logger

    @property
    def API_logger(self):
        return self.logger.getChild("api")

    @property
    def CLI_logger(self):
        return self.logger.getChild("cli")

    @property
    def config_logger(self):
        return self.logger.getChild("config")

    @property
    def runtime_info_logger(self):
        return self.logger.getChild("runtime")

    @property
    def is_config_loaded(self):
        return bool(self._config)

    @property
    def is_template_components_loaded(self):
        return bool(self._parameters)

    @property
    def config(self):
        if not self.is_config_loaded:
            self.load_config()
        return self._config

    def _load_config(self, config_dir, config_invocation_key, **overrides):
        self.logger.debug("Loading configuration.")
        self._config = Config(
            app=self,
            options=self.config_options,
            config_dir=config_dir,
            config_invocation_key=config_invocation_key,
            logger=self.config_logger,
            **overrides,
        )
        self.logger.info(f"Configuration loaded from: {self.config.config_file_path}")

    def load_config(self, config_dir=None, config_invocation_key=None, **overrides):
        if self.is_config_loaded:
            warnings.warn("Configuration is already loaded; reloading.")
        self._load_config(config_dir, config_invocation_key, **overrides)

    def reload_config(self, config_dir=None, config_invocation_key=None, **overrides):
        if not self.is_config_loaded:
            warnings.warn("Configuration is not loaded; loading.")
        self._load_config(config_dir, config_invocation_key, **overrides)

    def _make_API_CLI(self):
        """Generate the CLI for the main functionality."""

        @click.command(name="make")
        @click.argument("template_file_or_str")
        @click.option(
            "--string",
            is_flag=True,
            default=False,
            help="Determines if passing a file path or a string.",
        )
        @click.option(
            "--format",
            type=click.Choice(ALL_TEMPLATE_FORMATS),
            default=DEFAULT_TEMPLATE_FORMAT,
            help=(
                'If specified, one of "json" or "yaml". This forces parsing from a '
                "particular format."
            ),
        )
        @click.option(
            "--path",
            type=click.Path(exists=True),
            help="The directory path into which the new workflow will be generated.",
        )
        @click.option(
            "--name",
            help=(
                "The name of the workflow. If specified, the workflow directory will be "
                "`path` joined with `name`. If not specified the workflow template name "
                "will be used, in combination with a date-timestamp."
            ),
        )
        @click.option(
            "--overwrite",
            is_flag=True,
            default=False,
            help=(
                "If True and the workflow directory (`path` + `name`) already exists, "
                "the existing directory will be overwritten."
            ),
        )
        @click.option(
            "--store",
            type=click.Choice(ALL_STORE_FORMATS),
            help="The persistent store type to use.",
            default=DEFAULT_STORE_FORMAT,
        )
        @click.option(
            "--ts-fmt",
            help=(
                "The datetime format to use for storing datetimes. Datetimes are always "
                "stored in UTC (because Numpy does not store time zone info), so this "
                "should not include a time zone name."
            ),
        )
        @click.option(
            "--ts-name-fmt",
            help=(
                "The datetime format to use when generating the workflow name, where it "
                "includes a timestamp."
            ),
        )
        def make_workflow(
            template_file_or_str,
            string,
            format,
            path,
            name,
            overwrite,
            store,
            ts_fmt=None,
            ts_name_fmt=None,
        ):
            """Generate a new {app_name} workflow.

            TEMPLATE_FILE_OR_STR is either a path to a template file in YAML or JSON
            format, or a YAML/JSON string.

            """
            wk = self.make_workflow(
                template_file_or_str=template_file_or_str,
                is_string=string,
                template_format=format,
                path=path,
                name=name,
                overwrite=overwrite,
                store=store,
                ts_fmt=ts_fmt,
                ts_name_fmt=ts_name_fmt,
            )
            click.echo(wk.path)

        @click.command(name="go")
        @click.argument("template_file_or_str")
        @click.option(
            "--string",
            is_flag=True,
            default=False,
            help="Determines if passing a file path or a string.",
        )
        @click.option(
            "--format",
            type=click.Choice(ALL_TEMPLATE_FORMATS),
            default=DEFAULT_TEMPLATE_FORMAT,
            help=(
                'If specified, one of "json" or "yaml". This forces parsing from a '
                "particular format."
            ),
        )
        @click.option(
            "--path",
            type=click.Path(exists=True),
            help="The directory path into which the new workflow will be generated.",
        )
        @click.option(
            "--name",
            help=(
                "The name of the workflow. If specified, the workflow directory will be "
                "`path` joined with `name`. If not specified the workflow template name "
                "will be used, in combination with a date-timestamp."
            ),
        )
        @click.option(
            "--overwrite",
            is_flag=True,
            default=False,
            help=(
                "If True and the workflow directory (`path` + `name`) already exists, "
                "the existing directory will be overwritten."
            ),
        )
        @click.option(
            "--store",
            type=click.Choice(ALL_STORE_FORMATS),
            help="The persistent store type to use.",
            default=DEFAULT_STORE_FORMAT,
        )
        @click.option(
            "--ts-fmt",
            help=(
                "The datetime format to use for storing datetimes. Datetimes are always "
                "stored in UTC (because Numpy does not store time zone info), so this "
                "should not include a time zone name."
            ),
        )
        @click.option(
            "--ts-name-fmt",
            help=(
                "The datetime format to use when generating the workflow name, where it "
                "includes a timestamp."
            ),
        )
        @click.option(
            "--js-parallelism",
            help=(
                "If True, allow multiple jobscripts to execute simultaneously. Raises if "
                "set to True but the store type does not support the "
                "`jobscript_parallelism` feature. If not set, jobscript parallelism will "
                "be used if the store type supports it."
            ),
            type=click.BOOL,
        )
        def make_and_submit_workflow(
            template_file_or_str,
            string,
            format,
            path,
            name,
            overwrite,
            store,
            ts_fmt=None,
            ts_name_fmt=None,
            js_parallelism=None,
        ):
            """Generate and submit a new {app_name} workflow.

            TEMPLATE_FILE_OR_STR is either a path to a template file in YAML or JSON
            format, or a YAML/JSON string.

            """
            self.make_and_submit_workflow(
                template_file_or_str=template_file_or_str,
                is_string=string,
                template_format=format,
                path=path,
                name=name,
                overwrite=overwrite,
                store=store,
                ts_fmt=ts_fmt,
                ts_name_fmt=ts_name_fmt,
                JS_parallelism=js_parallelism,
            )

        @click.command(context_settings={"ignore_unknown_options": True})
        @click.argument("py_test_args", nargs=-1, type=click.UNPROCESSED)
        @click.pass_context
        def test(ctx, py_test_args):
            """Run {app_name} test suite.

            PY_TEST_ARGS are arguments passed on to Pytest.

            """
            ctx.exit(self.run_tests(*py_test_args))

        @click.command(context_settings={"ignore_unknown_options": True})
        @click.argument("py_test_args", nargs=-1, type=click.UNPROCESSED)
        @click.pass_context
        def test_hpcflow(ctx, py_test_args):
            """Run hpcflow test suite.".

            PY_TEST_ARGS are arguments passed on to Pytest.

            """
            ctx.exit(self.run_hpcflow_tests(*py_test_args))

        commands = [
            make_workflow,
            make_and_submit_workflow,
            test,
        ]
        for cmd in commands:
            if cmd.help:
                cmd.help = cmd.help.format(app_name=self.name)

        if type(self) is not BaseApp:
            # `test_hpcflow` is the same as `test` for the BaseApp so no need to add both:
            commands.append(test_hpcflow)

        return commands

    def _make_workflow_submission_jobscript_CLI(self):
        """Generate the CLI for interacting with existing workflow submission
        jobscripts."""

        @click.group(name="js")
        @click.pass_context
        @click.argument("js_idx", type=click.INT)
        def jobscript(ctx, js_idx):
            """Interact with existing {app_name} workflow submission jobscripts.

            JS_IDX is the jobscript index within the submission object.

            """
            ctx.obj["jobscript"] = ctx.obj["submission"].jobscripts[js_idx]

        @jobscript.command(name="res")
        @click.pass_context
        def resources(ctx):
            """Get resources associated with this jobscript."""
            click.echo(ctx.obj["jobscript"].resources.__dict__)

        @jobscript.command(name="deps")
        @click.pass_context
        def dependencies(ctx):
            """Get jobscript dependencies."""
            click.echo(ctx.obj["jobscript"].dependencies)

        @jobscript.command()
        @click.pass_context
        def path(ctx):
            """Get the file path to the jobscript."""
            click.echo(ctx.obj["jobscript"].jobscript_path)

        @jobscript.command()
        @click.pass_context
        def show(ctx):
            """Show the jobscript file."""
            with ctx.obj["jobscript"].jobscript_path.open("rt") as fp:
                click.echo(fp.read())

        jobscript.help = jobscript.help.format(app_name=self.name)

        return jobscript

    def _make_workflow_submission_CLI(self):
        """Generate the CLI for interacting with existing workflow submissions."""

        @click.group(name="sub")
        @click.pass_context
        @click.argument("sub_idx", type=click.INT)
        def submission(ctx, sub_idx):
            """Interact with existing {app_name} workflow submissions.

            SUB_IDX is the submission index.

            """
            ctx.obj["submission"] = ctx.obj["workflow"].submissions[sub_idx]

        @submission.command("status")
        @click.pass_context
        def status(ctx):
            """Get the submission status."""
            click.echo(ctx.obj["submission"].status.name.lower())

        @submission.command("submitted-js")
        @click.pass_context
        def submitted_JS(ctx):
            """Get a list of jobscript indices that have been submitted."""
            click.echo(ctx.obj["submission"].submitted_jobscripts)

        @submission.command("outstanding-js")
        @click.pass_context
        def outstanding_JS(ctx):
            """Get a list of jobscript indices that have not yet been submitted."""
            click.echo(ctx.obj["submission"].outstanding_jobscripts)

        @submission.command("needs-submit")
        @click.pass_context
        def needs_submit(ctx):
            """Check if this submission needs submitting."""
            click.echo(ctx.obj["submission"].needs_submit)

        submission.help = submission.help.format(app_name=self.name)
        submission.add_command(self._make_workflow_submission_jobscript_CLI())

        return submission

    def _make_workflow_CLI(self):
        """Generate the CLI for interacting with existing workflows."""

        @click.group()
        @click.argument("workflow_path", type=click.Path(exists=True))
        @click.pass_context
        def workflow(ctx, workflow_path):
            """Interact with existing {app_name} workflows.

            WORKFLOW_PATH is the path to an existing workflow.

            """
            wk = self.Workflow(workflow_path)
            ctx.ensure_object(dict)
            ctx.obj["workflow"] = wk

        @workflow.command(name="submit")
        @click.option(
            "--js-parallelism",
            help=(
                "If True, allow multiple jobscripts to execute simultaneously. Raises if "
                "set to True but the store type does not support the "
                "`jobscript_parallelism` feature. If not set, jobscript parallelism will "
                "be used if the store type supports it."
            ),
            type=click.BOOL,
        )
        @click.pass_context
        def submit_workflow(ctx, js_parallelism=None):
            """Submit the workflow."""
            ctx.obj["workflow"].submit(JS_parallelism=js_parallelism)

        @workflow.command(name="get-param")
        @click.argument("index", type=click.INT)
        @click.pass_context
        def get_parameter(ctx, index):
            """Get a parameter value by data index."""
            click.echo(ctx.obj["workflow"].get_parameter_data(index))

        @workflow.command(name="get-param-source")
        @click.argument("index", type=click.INT)
        @click.pass_context
        def get_parameter_source(ctx, index):
            """Get a parameter source by data index."""
            click.echo(ctx.obj["workflow"].get_parameter_source(index))

        @workflow.command(name="get-all-params")
        @click.pass_context
        def get_all_parameters(ctx):
            """Get all parameter values."""
            click.echo(ctx.obj["workflow"].get_all_parameter_data())

        @workflow.command(name="is-param-set")
        @click.argument("index", type=click.INT)
        @click.pass_context
        def is_parameter_set(ctx, index):
            """Check if a parameter specified by data index is set."""
            click.echo(ctx.obj["workflow"].is_parameter_set(index))

        @workflow.command(name="show-all-status")
        @click.pass_context
        def show_all_EAR_statuses(ctx):
            """Show the submission status of all workflow EARs."""
            ctx.obj["workflow"].show_all_EAR_statuses()

        workflow.help = workflow.help.format(app_name=self.name)

        workflow.add_command(self._make_workflow_submission_CLI())

        return workflow

    def _make_submission_CLI(self):
        """Generate the CLI for submission related queries."""

        def OS_info_callback(ctx, param, value):
            if not value or ctx.resilient_parsing:
                return
            click.echo(self.get_OS_info())
            ctx.exit()

        @click.group()
        @click.option(
            "--os-info",
            help="Print information about the operating system.",
            is_flag=True,
            is_eager=True,
            expose_value=False,
            callback=OS_info_callback,
        )
        @click.pass_context
        def submission(ctx):
            """Submission-related queries."""

        @submission.command("shell-info")
        @click.argument("shell_name", type=click.Choice(ALL_SHELLS))
        @click.option("--exclude-os", is_flag=True, default=False)
        @click.pass_context
        def shell_info(ctx, shell_name, exclude_os):
            click.echo(self.get_shell_info(shell_name, exclude_os))
            ctx.exit()

        return submission

    def _make_internal_CLI(self):
        """Generate the CLI for internal use."""

        @click.group()
        def internal(help=True):  # TEMP
            """Internal CLI to be invoked by scripts generated by the app."""
            pass

        @internal.group()
        @click.argument("path", type=click.Path(exists=True))
        @click.pass_context
        def workflow(ctx, path):
            """"""
            wk = self.Workflow(path)
            ctx.ensure_object(dict)
            ctx.obj["workflow"] = wk

        @workflow.command()
        @click.pass_context
        @click.argument("submission_idx", type=click.INT)
        @click.argument("jobscript_idx", type=click.INT)
        @click.argument("js_element_idx", type=click.INT)
        @click.argument("js_action_idx", type=click.INT)
        def write_commands(
            ctx,
            submission_idx: int,
            jobscript_idx: int,
            js_element_idx: int,
            js_action_idx: int,
        ):
            ctx.exit(
                ctx.obj["workflow"].write_commands(
                    submission_idx,
                    jobscript_idx,
                    js_element_idx,
                    js_action_idx,
                )
            )

        @workflow.command()
        @click.pass_context
        @click.argument("name")
        @click.argument("value")
        @click.argument("submission_idx", type=click.INT)
        @click.argument("jobscript_idx", type=click.INT)
        @click.argument("js_element_idx", type=click.INT)
        @click.argument("js_action_idx", type=click.INT)
        def save_parameter(
            ctx,
            name: str,
            value: str,
            submission_idx: int,
            jobscript_idx: int,
            js_element_idx: int,
            js_action_idx: int,
        ):
            ctx.exit(
                ctx.obj["workflow"].save_parameter(
                    name,
                    value,
                    submission_idx,
                    jobscript_idx,
                    js_element_idx,
                    js_action_idx,
                )
            )

        @workflow.command()
        @click.pass_context
        @click.argument("submission_idx", type=click.INT)
        @click.argument("jobscript_idx", type=click.INT)
        @click.argument("js_element_idx", type=click.INT)
        @click.argument("js_action_idx", type=click.INT)
        def set_EAR_start(
            ctx,
            submission_idx: int,
            jobscript_idx: int,
            js_element_idx: int,
            js_action_idx: int,
        ):
            ctx.exit(
                ctx.obj["workflow"].set_EAR_start(
                    submission_idx,
                    jobscript_idx,
                    js_element_idx,
                    js_action_idx,
                )
            )

        @workflow.command()
        @click.pass_context
        @click.argument("submission_idx", type=click.INT)
        @click.argument("jobscript_idx", type=click.INT)
        @click.argument("js_element_idx", type=click.INT)
        @click.argument("js_action_idx", type=click.INT)
        def set_EAR_end(
            ctx,
            submission_idx: int,
            jobscript_idx: int,
            js_element_idx: int,
            js_action_idx: int,
        ):
            ctx.exit(
                ctx.obj["workflow"].set_EAR_end(
                    submission_idx,
                    jobscript_idx,
                    js_element_idx,
                    js_action_idx,
                )
            )

        # TODO: in general, maybe the workflow command group can expose the simple Workflow
        # properties; maybe use a decorator on the Workflow property object to signify
        # inclusion?

        return internal

    def _make_CLI(self):
        """Generate the root CLI for the app."""

        colorama_init(autoreset=True)

        def run_time_info_callback(ctx, param, value):
            if not value or ctx.resilient_parsing:
                return
            click.echo(str(self.run_time_info))
            ctx.exit()

        @click.group(name=self.name)
        @click.version_option(
            version=self.version,
            package_name=self.name,
            prog_name=self.name,
            help=f"Show the version of {self.name} and exit.",
        )
        @click.version_option(
            __version__,
            "--hpcflow-version",
            help="Show the version of hpcflow and exit.",
            package_name="hpcflow",
            prog_name="hpcflow",
        )
        @click.help_option()
        @click.option(
            "--run-time-info",
            help="Print run-time information!",
            is_flag=True,
            is_eager=True,
            expose_value=False,
            callback=run_time_info_callback,
        )
        @click.option("--config-dir", help="Set the configuration directory.")
        @click.option(
            "--config-invocation-key", help="Set the configuration invocation key."
        )
        @click.option(
            "--with-config",
            help="Override a config item in the config file",
            nargs=2,
            multiple=True,
        )
        @click.pass_context
        def new_CLI(ctx, config_dir, config_invocation_key, with_config):
            overrides = {kv[0]: kv[1] for kv in with_config}
            try:
                self.load_config(
                    config_dir=config_dir,
                    config_invocation_key=config_invocation_key,
                    **overrides,
                )
            except ConfigError as err:
                click.echo(f"{colored(err.__class__.__name__, 'red')}: {err}")
                ctx.exit(1)

        new_CLI.__doc__ = self.description
        new_CLI.add_command(get_config_CLI(self))
        new_CLI.add_command(get_demo_software_CLI(self))
        new_CLI.add_command(get_helper_CLI(self))
        new_CLI.add_command(self._make_workflow_CLI())
        new_CLI.add_command(self._make_submission_CLI())
        new_CLI.add_command(self._make_internal_CLI())
        for cli_cmd in self._make_API_CLI():
            new_CLI.add_command(cli_cmd)

        return new_CLI

    def _load_scripts(self):
        # TODO: load custom directories / custom functions (via decorator)
        pkg = "hpcflow.sdk.demo.scripts"
        script_names = (
            name
            for name in resources.contents(pkg)
            if name != "__init__.py" and resources.is_resource(pkg, name)
        )
        scripts = {}
        for i in script_names:
            scripts[i] = resources.path(pkg, i)
        return scripts

    def template_components_from_json_like(self, json_like):
        cls_lookup = {
            "parameters": self.ParametersList,
            "command_files": self.CommandFilesList,
            "environments": self.EnvironmentsList,
            "task_schemas": self.TaskSchemasList,
        }
        tc = {}
        for k, v in cls_lookup.items():
            tc_k = v.from_json_like(
                json_like.get(k, {}),
                shared_data=tc,
                is_hashed=True,
            )
            tc[k] = tc_k
        return tc

    def get_info(self) -> Dict:
        return {
            "name": self.name,
            "version": self.version,
            "python_version": self.run_time_info.python_version,
            "is_frozen": self.run_time_info.is_frozen,
        }


class App(BaseApp):
    """Class to generate an hpcflow application (e.g. MatFlow)"""

    pass



# from hpcflow.parameters import InputValue, Parameter
# from hpcflow.task import TaskTemplate
# from hpcflow.task_schema import TaskObjective, TaskSchema


# def test_expected_return_resolve_elements():

#     generate_RVE = TaskObjective("generate_volume_element")
#     simulate = TaskObjective("simulate_volume_element_loading")

#     grid_size = Parameter("grid_size")
#     size = Parameter("size")
#     volume_element = Parameter("volume_element")
#     load_case = Parameter("load_case")
#     volume_element_response = Parameter("volume_element_response")

#     generate_RVE_schema = TaskSchema(
#         objective=generate_RVE, inputs=[grid_size, size], outputs=[volume_element]
#     )
#     simulate_schema = TaskSchema(
#         objective=simulate,
#         inputs=[volume_element, load_case],
#         outputs=[volume_element_response],
#     )

#     task_1 = TaskTemplate(
#         schema=generate_RVE_schema,
#         input_values=[
#             InputValue(grid_size, value=[8, 8, 8]),
#             InputValue(size, value=[1, 1, 1]),
#         ],
#     )
#     task_2 = TaskTemplate(
#         schema=simulate_schema, input_values=[InputValue(load_case, value=10)]
#     )



from dataclasses import dataclass
import re
from typing import List, Any

from hpcflow.sdk.core.json_like import JSONLike


@dataclass
class Command(JSONLike):

    _app_attr = "app"

    command: str
    arguments: List[Any] = None
    stdout: str = None
    stderr: str = None
    stdin: str = None

    def __repr__(self) -> str:
        out = []
        if self.command:
            out.append(f"command={self.command!r}")
        if self.arguments:
            out.append(f"arguments={self.arguments!r}")
        if self.stdout:
            out.append(f"stdout={self.stdout!r}")
        if self.stderr:
            out.append(f"stderr={self.stderr!r}")
        if self.stdin:
            out.append(f"stdin={self.stdin!r}")

        return f"{self.__class__.__name__}({', '.join(out)})"

    def get_output_types(self):
        # note: we use "parameter" rather than "output", because it could be a schema
        # output or schema input.
        vars_regex = r"\<\<parameter:(.*?)\>\>"
        out = {"stdout": None, "stderr": None}
        for i, label in zip((self.stdout, self.stderr), ("stdout", "stderr")):
            if i:
                match = re.search(vars_regex, i)
                if match:
                    param_typ = match.group(1)
                    if match.span(0) != (0, len(i)):
                        raise ValueError(
                            f"If specified as a parameter, `{label}` must not include"
                            f" any characters other than the parameter "
                            f"specification, but this was given: {i!r}."
                        )
                    out[label] = param_typ
        return out


@dataclass
class CommandArgument:
    """
    Attributes
    ----------
    parts : list of any of str, File, Parameter

    """

    parts: List[Any]



from hpcflow import __version__
from hpcflow.sdk import ConfigOptions
from hpcflow.sdk.app import BaseApp

config_options = ConfigOptions(
    directory_env_var="HPCFLOW_CONFIG_DIR",
    default_directory="~/.hpcflow",
    sentry_DSN="https://2463b288fd1a40f4bada9f5ff53f6811@o1180430.ingest.sentry.io/6293231",
    sentry_traces_sample_rate=1.0,
    sentry_env="main" if "a" in __version__ else "develop",
)

# built in template components (in this case, for demonstration purposes):
template_components = BaseApp.load_builtin_template_component_data(
    "hpcflow.sdk.data.template_components"
)

hpcflow = BaseApp(
    name="hpcflow",
    version=__version__,
    description="Computational workflow management",
    config_options=config_options,
    template_components=template_components,
    pytest_args=[
        "--verbose",
        "--exitfirst",
    ],
)

load_config = hpcflow.load_config
reload_config = hpcflow.reload_config
make_workflow = hpcflow.make_workflow

# expose core classes that require access to the App instance:
Action = hpcflow.Action
ActionEnvironment = hpcflow.ActionEnvironment
ActionScope = hpcflow.ActionScope
ActionScopeType = hpcflow.ActionScopeType
ActionRule = hpcflow.ActionRule
Command = hpcflow.Command
Environment = hpcflow.Environment
Executable = hpcflow.Executable
ExecutableInstance = hpcflow.ExecutableInstance
ExecutablesList = hpcflow.ExecutablesList
ElementPropagation = hpcflow.ElementPropagation
ElementSet = hpcflow.ElementSet
FileSpec = hpcflow.FileSpec
InputFile = hpcflow.InputFile
InputFileGenerator = hpcflow.InputFileGenerator
InputSource = hpcflow.InputSource
InputSourceType = hpcflow.InputSourceType
InputValue = hpcflow.InputValue
Loop = hpcflow.Loop
OutputFileParser = hpcflow.OutputFileParser
Parameter = hpcflow.Parameter
ParameterValue = hpcflow.ParameterValue
ResourceList = hpcflow.ResourceList
ResourceSpec = hpcflow.ResourceSpec
SchemaInput = hpcflow.SchemaInput
SchemaOutput = hpcflow.SchemaOutput
Task = hpcflow.Task
TaskObjective = hpcflow.TaskObjective
TaskSchema = hpcflow.TaskSchema
TaskSourceType = hpcflow.TaskSourceType
ValueSequence = hpcflow.ValueSequence
Workflow = hpcflow.Workflow
WorkflowTask = hpcflow.WorkflowTask
WorkflowTemplate = hpcflow.WorkflowTemplate
